import datetime
import itertools

import math

import numpy as np
import sys

from csm.calculations.basic_calculations import check_perm_cycles, now, run_time
from csm.calculations.data_classes import CSMState, Operation, CSMResult
from csm.calculations.constants import MINDOUBLE, MAXDOUBLE
from csm.fast import calc_ref_plane
from csm.fast import CythonPermuter, SinglePermPermuter
import logging

from csm.input_output.formatters import format_perm_count
from csm.input_output.formatters import csm_log as print
np.set_printoptions(precision=6)


__author__ = 'Itay, Devora, Yael'



# When this property is set by an outside caller, it is called every permutation iteration with the current CSMState
# This is useful for writing all permutations to file during the calculation
csm_state_tracer_func = None

class CSMValueError(ValueError):
    def __init__(self, arg1, CSMState):
        self.arg1 = arg1
        self.CSMState = CSMState
        super().__init__(arg1)


class ExactStatistics:
    def __init__(self, permuter):
        self._perm_count=permuter.count
        self._truecount=permuter.truecount
        self._falsecount=permuter.falsecount

    def write(self, f=sys.stderr):
        f.write("Number of permutations: %s" % format_perm_count(self.perm_count))
        f.write("Number of branches in permutation tree: %s" % format_perm_count(self.num_branches))
        f.write("Number of dead ends: %s" % format_perm_count(self.dead_ends))

    def to_dict(self):
        return {
            "perm count":self.perm_count,
            "number branches":self.num_branches,
            "dead ends":self.dead_ends
        }

    @property
    def dead_ends(self):
        return self._falsecount

    @property
    def perm_count(self):
        return self._perm_count

    @property
    def num_branches(self):
        return self._truecount

class ExactCalculation:
    def __init__(self, operation, molecule, *args, **kwargs):
        """
        A class for running the exact CSM Algorithm
        :param operation: instance of Operation class or named tuple, with fields for name and order, that describes the symmetry
        :param molecule: instance of Molecule class on which the described symmetry calculation will be performed
        :param keep_structure: boolean, when True only permutations that maintain bond integrity will be measured 
        :param perm: a list of atom indiced describing one permutation of the molecule. Default None- When provided,
         only the provided permutation is measured
        :param no_constraint: boolean, default False, when False the constraints algorithm is used for the permuter, 
        when True the old permuter is used
        :param timeout: default 300, the number of seconds the function will run before timing out
        :param callback_func: default None, this function is called for every single permutation calculated with an argument of a single 
        CSMState, can be used for printing in-progress reports, outputting to an excel, etc.
        """
        self.operation=operation
        self.molecule=molecule

    def calculate(self, timeout=300, *args, **kwargs):
        self.start_time = now()
        op_type=self.operation.type
        op_order=self.operation.order
        molecule=self.molecule


        if op_type == 'CH':  # Chirality
            # sn_max = op_order
            # First CS
            best_result = self.csm_operation('CS', 2, molecule, timeout=timeout)
            best_result = best_result._replace(op_type='CS')  # unclear why this line isn't redundant
            if best_result.csm > MINDOUBLE:
                # Try the SN's
                for op_order in range(2, self.operation.order + 1, 2):
                    result = self.csm_operation('SN', op_order, molecule, timeout=timeout)
                    if result.csm < best_result.csm:
                        best_result = result._replace(op_type='SN', op_order=op_order)
                    if best_result.csm < MINDOUBLE:
                        break

        else:
            best_result = self.csm_operation(op_type, op_order, molecule, timeout=timeout)

        overall_stats = self.statistics.to_dict()
        overall_stats["runtime"]=run_time(self.start_time)
        self._csm_result = CSMResult(best_result, self.operation, overall_stats=overall_stats)
        return self.result

    def csm_operation(self, op_type, op_order, molecule, perm=None, timeout=300):
        """
        Calculates minimal csm, directional cosines by applying permutations that keep the similar atoms within the group.
        :param op_type: cannot be CH.
        :param op_order:
        :param molecule:
        :param keep_structure:
        :param perm:
        :param no_constraint:
        :param suppress_print:
        :param timeout:
        :return:
        """
        best_csm = CSMState(molecule=molecule, op_type=op_type, op_order=op_order, csm=MAXDOUBLE)
        traced_state = CSMState(molecule=molecule, op_type=op_type, op_order=op_order)

        if perm:
            permuter = SinglePermPermuter(np.array(perm, dtype="long"), molecule, op_order, op_type)
        else:
            permuter = CythonPermuter(molecule, op_order, op_type, timeout=timeout)

        for calc_state in permuter.permute():
            if permuter.count % 1000000 == 0:
                print("calculated for", int(permuter.count / 1000000), "million permutations thus far...\t Time:",
                      run_time(self.start_time))
            csm, dir = calc_ref_plane(op_order, op_type == 'CS', calc_state)

            if csm < best_csm.csm:
                best_csm = best_csm._replace(csm=csm, dir=dir, perm=list(calc_state.perm))

        self.statistics=ExactStatistics(permuter)

        if best_csm.csm == MAXDOUBLE:
            # failed to find csm value for any permutation
            #best_csm = best_csm._replace(csm=csm, dir=dir, perm=list(calc_state.perm))
            raise CSMValueError("Failed to calculate a csm value for %s %d" % (op_type, op_order), best_csm)
        return best_csm

    @staticmethod
    def exact_calculation_for_approx(operation, molecule, perm):
        ec = ExactCalculation(operation, molecule)
        if operation.type == 'CH':  # Chirality
            best_result = ec.csm_operation('CS', 2, molecule,perm=perm)
            if best_result.csm > MINDOUBLE:
                # Try the SN's
                for op_order in range(2, operation.order + 1, 2):
                    result = ec.csm_operation('SN', op_order, molecule,perm=perm)
                    if result.csm < best_result.csm:
                        best_result = result._replace(op_type='SN', op_order=op_order)
                    if best_result.csm < MINDOUBLE:
                        break
        else:
            best_result=ec.csm_operation(operation.type, operation.order, molecule, perm=perm)

        falsecount, num_invalid, cycle_counts, bad_indices = check_perm_cycles(perm, operation)
        best_result=best_result._replace(num_invalid=num_invalid)
        return best_result

    @property
    def result(self):
        return self._csm_result







import numpy as np
import math as m
from datetime import datetime




def now():
    return datetime.now()

def run_time(start_time):
    now = datetime.now()
    time_d = now - start_time
    return time_d.total_seconds()

class CalculationTimeoutError(TimeoutError):
    def __init__(self, timeout_delta, *args, **kwargs):
        super().__init__("Calculation timed out after "+str(timeout_delta)+" seconds", *args, **kwargs)
        self.timeout_delta=timeout_delta

def check_timeout(local_start, local_timeout):
    from csm.calculations.constants import global_start_time, global_time_out
    runtime=run_time(local_start)
    if runtime>local_timeout:
        raise CalculationTimeoutError(runtime)
    g_runtime=run_time(global_start_time)
    if g_runtime>global_time_out:
        raise CalculationTimeoutError(g_runtime)


def cart2sph(x,y,z, normalize=True):
    def do_normalize(x, y, z):
        norm= x*x + y*y + z*z
        return x/norm, y/norm, z/norm
    #https://stackoverflow.com/questions/4116658/faster-numpy-cartesian-to-spherical-coordinate-conversion
    if normalize:
        x, y, z = do_normalize(x, y, z)
    XsqPlusYsq = x**2 + y**2
    r = m.sqrt(XsqPlusYsq + z**2)               # r
    elev = m.atan2(z,m.sqrt(XsqPlusYsq))     # theta
    az = m.atan2(y,x)                           # phi
    return r, elev, az




def create_rotation_matrix(iOp, op_type, op_order, dir):
    is_improper = op_type != 'CN'
    is_zero_angle = op_type == 'CS'
    W = np.array([[0.0, -dir[2], dir[1]], [dir[2], 0.0, -dir[0]], [-dir[1], dir[0], 0.0]])
    rot = np.zeros((3, 3))
    angle = 0.0 if is_zero_angle else 2 * np.pi * iOp / op_order
    factor = -1 if is_improper and (iOp % 2) == 1 else 1

    # The rotation matrix is calculated similarly to the Rodrigues rotation matrix. The only
    # difference is that the matrix is also a reflection matrix when factor is -1.
    #
    # This is why we took the old C++ code instead of applying the Rodrigues formula directly.
    for s in range(3):
        for t in range(3):
            ang = np.cos(angle) if s == t else 0
            rot[s][t] = ang + ((factor - np.cos(angle)) * dir[s] * dir[t] + np.sin(angle) * W[s][t])

    return rot






def check_perm_cycles(perm, operation):
    '''
    This function checks the cycles in a given permutation according to the provided operation order and type.
    It counts the legal and illegal cycles in the permutation
    :param perm:
    :param op_order: 
    :param op_type: 
    :return: the number of illegal cycles, the number of molecules in illegal cycles, a dictionary of cycle lengths-
    with key =length cycle, val= mnumber of cycles of that length, and an array of the indices in bad cycles
    '''
    op_order=operation.order
    op_type=operation.type
    checked=[False]*len(perm)
    num_molecules_in_bad_cycles=0
    num_good_cycles=0
    num_bad_cycles=0

    cycle_counts={}

    indices_in_bad_cycles=[]


    for i, index in enumerate(perm):
        if checked[i]:
            continue
        checked[i]=True
        cycle=[index]
        while not checked[index]:
            checked[index] = True
            index = perm[index]
            cycle.append(index)

        cycle_len=len(cycle)

        if cycle_len in cycle_counts:
            cycle_counts[cycle_len]+=1
        else:
            cycle_counts[cycle_len]=1

        if cycle_len == 1 or cycle_len == op_order or (cycle_len == 2 and op_type == 'SN'):
            num_good_cycles += 1
        else:
            num_molecules_in_bad_cycles += cycle_len
            num_bad_cycles+=1
            indices_in_bad_cycles+=cycle


    return num_bad_cycles, num_molecules_in_bad_cycles, cycle_counts, indices_in_bad_cycles






def check_perm_equivalence(mol, perm):
    for origin, destination in enumerate(perm):
        if destination not in mol.atoms[origin].equivalency:
            return False
    return True

def check_perm_structure_preservation(mol, perm):
    '''
    checks what percent of a permutation does not break bonds in the molecule--
    so a return value of 1 would be a perfectly preserved bond structure, of 0.5 would mean half of the molecules bonds are broken
    :param mol: the molecule being permuted
    :param perm: the permutation whose structure preservation is being measured
    :return: percent of bonds that are preserved
    '''
    if len(mol.bondset)==0:
        raise ValueError("Molecule does not have any bond information")

    broken=0
    for origin, destination in enumerate(perm):
        for adjacent in mol.atoms[origin].adjacent:
            if (destination, perm[adjacent]) not in mol.bondset:
                broken+=1

    percent_structure= (len(mol.bondset)- broken )/len(mol.bondset)

    return percent_structure


def array_distance(a, b):
    return np.sqrt(
        (a[0] - b[0]) * (a[0] - b[0])
        + (a[1] - b[1]) * (a[1] - b[1])
        + (a[2] - b[2]) * (a[2] - b[2]))


import math
import sys

def format_perm_count(num):
    if abs(num) < 1000000:
        return '%d' % num
    return '%.5g' % num

def format_CSM(num):
    return "%.4lf" % (abs(num))



def non_negative_zero(number):
    if math.fabs(number)<0.00001:
        return 0.0000
    else:
        return number

def format_unknown_str(input):
    try:
        num=float(input)
        if int(input)==num:
            return int(input)
        return "%.4lf" % (num)
    except:
        return str(input)

global csm_out_pipe
csm_out_pipe=sys.stdout

global output_strings
output_strings=[]

def csm_log(*strings, file=csm_out_pipe, **kwargs):
    output_strings.append(" ".join(str(s) for s in strings))
    print(*strings, **kwargs, file=file)



from csm.calculations.constants import MINDOUBLE

__author__ = 'zmbq'

import math
import logging
logger = logging.getLogger("csm")


def calculate_norm_factor(coords, center_of_mass):
    """
    :param coords: list of atom coordinates
    :param center_of_mass: center of mass of molecule
    """
    size=len(coords)
    norm = 0.0
    for i in range(size):
        tmp = (coords[i][0] - center_of_mass[0]) ** 2 + (coords[i][1] - center_of_mass[1]) ** 2 + (coords[i][2] - center_of_mass[2]) ** 2
        norm += tmp
        #logger.debug("Norm: %lf i: %lf temp %lf" % (norm, i, tmp))

    # normalize to 1 and not molecule size
    # norm = sqrt(norm / (double)m->size());

    norm = math.sqrt(norm)
    #logger.debug("Second normalization factor is %lf and average is (%lf, %lf, %lf)" % (norm, center_of_mass[0], center_of_mass[1], center_of_mass[2]))

    if norm<=MINDOUBLE: #in the original code, this check was against MINDOUBLE.
        raise(ValueError("Normalization factor equals zero"))
        #norm=default_value

    return norm

def normalize_coords(coords, center_of_mass, norm_factor):
    """
    :param coords: list of atom coordinates
    :param center_of_mass: center of mass of molecule
    :param norm_factor: nromalization factor (can be calculated with calculate_norm_factor)
    :return: list of atom coordinates, normalized and moved to origin
    """

    for i in range(len(coords)):
        coords[i] = ((coords[i][0] - center_of_mass[0]) / norm_factor,
                     (coords[i][1] - center_of_mass[1]) / norm_factor,
                     (coords[i][2] - center_of_mass[2]) / norm_factor)


    return coords


def de_normalize_coords(coords, norm_factor):
    size = len(coords)
    for i in range(size):
        coords[i] = (coords[i][0] * norm_factor,
                     coords[i][1] * norm_factor,
                     coords[i][2] * norm_factor)
    return coords




import sys
from csm.molecule.molecule import MoleculeReader
from csm.calculations.basic_calculations import check_perm_structure_preservation, check_perm_equivalence, check_perm_cycles
from csm.input_output.formatters import csm_log as print




def check_perm_validity(mol, perm, **kwargs):
    '''
    Checks whether a user-input permutation is valid-- has legal cycle lengths, only switches between equivalence classes,
    and maintains molecule bonds. Warns for each violation.
    :param mol: the molecule being permuted
    :param perm: the permutation
    '''
    falsecount, num_invalid, cycle_counts, bad_indices= check_perm_cycles(perm, kwargs['operation'])
    if falsecount>0:
        print("Warning: Permutation does not maintain cycle structure")
    if not check_perm_equivalence(mol, perm):
        print("Warning: Permutation contains switches between non-equivalent atoms")
    try:
        if check_perm_structure_preservation(mol, perm) < 1:
            print("Warning: Permutation does not preserve molecule structure")
    except ValueError: #molecule has no structure
        pass


def read_perm_file(filename):
    """
    Reads a permutation
    :param filename: Name of perm file
    :return: permutation as a list of numbers

    Check that the permutation is legal, raise ValueError if not
    """
    with open(filename, 'r') as f:
        line = f.read().split()
        used = set()

        result = []
        for num_str in line:
            try:
                num = int(num_str)
            except ValueError:
                raise ValueError("Invalid permutation %s - only numbers are allowed" % line)
            if num < 1 or num > len(line):
                raise ValueError("Invalid permutation %s - out of range" % line)
            num -= 1
            if num  in used:
                raise ValueError("Invalid permutation %s - number appears twice" % line)
            result.append(num)
            used.add(num)

    return result


def read_perm(molecule, perm_file_name=None, **kwargs):
    '''
    :param molecule: the molecule whose permutation is being read (necessary for validity checks)
    :param perm_file_name: the name of the file where the permutation is stored (file permutation is written in indexes starting from 1)
    :return: permutation- a list of permuted indexes (starting with 0) (can be None)
    '''
    if perm_file_name:
        perm = read_perm_file(perm_file_name)
        if len(perm) != len(molecule):
            raise ValueError("Invalid permutation - permutation is of size %d but molecule has %d atoms" %
                             (len(perm), len(molecule)))
        check_perm_validity(molecule, perm, **kwargs)
    else:
        perm = None
    return perm


def read_dir_file(filename):
    """
    THIS FUNCTION HAS BEEN RETIRED AND IS NOT IN USE AT THE MOMENT
    Reads a symmetry direction file
    :param filename: Name of dir file
    :return: (x,y,z) of the symmetry axis
    """
    dirs=[]
    try:
        with open(filename, 'r') as f:
            for fileline in f:
                line = fileline.split()
                result = (float(line[0]), float(line[1]), float(line[2]))
                dirs.append(result)
        if len(dirs)==0:
            raise ValueError("provided dir file is empty")
        return dirs
    except Exception as e:
        raise ValueError("Invalid input for use-dir")





def read_from_sys_std_in():
    '''
    a wrapper around calls to sys.stdin.read: we first check that there's something being piped, and if not, raise an error
    (otherwise the program hands forever)
    :return:
    '''
    if not sys.stdin.isatty():
        raw_json = sys.stdin.read()
        return raw_json
    else:
        raise ValueError("No input in sys.stdin")


from collections import namedtuple
from openbabel import OBAtom, OBElementTable

__author__ = 'zmbq'

_tbl = OBElementTable()

ChainedPermutation = namedtuple('ChainedPermutation', ['chain_perm', 'atom_perm'])

def GetAtomicMass(symbol):
    """ Returns the atomic mass of an element
    :param symbol: Element's symbol
    :return: The atomic mass
    """
    atomicNum = _tbl.GetAtomicNum(symbol)
    atom = OBAtom()
    atom.SetAtomicNum(atomicNum)
    atom.SetIsotope(0)
    return atom.GetAtomicMass()


def GetAtomicSymbol(atomic_num):
    """ Returns the element symbol from the atomic number
    :param atomic_num: Atomic number
    :return: Element symbol
    """
    return _tbl.GetSymbol(atomic_num)


class Atom:
    """ A single atom, alogn with its position and neighbors
    """
    def __init__(self, symbol, pos, index, useMass=False, chain='Simulated Chain'):
        """
        :param symbol: atomic symbol (ie 'C', 'H', 'N')
        :param pos: position coordinates as a list of floats [1.0, 0.0, 0.0]
        :param index: integer index in molecule list of atoms
        :param useMass: boolean, when True atomic mass is used, when False mass=1.0
        :param chain: string, name of the chain the atom belongs to
        """
        self.index=index
        self._symbol=symbol.strip() #make sure no white space in symbol
        self.adjacent = []
        self.pos = pos
        if useMass and symbol != 'XX':
            self._mass = GetAtomicMass(symbol)
        else:
            self._mass = 1.0
        self._chain = chain
        self._equivalency=[]

    @property
    def mass(self):
        return self._mass

    @property
    def symbol(self):
        return self._symbol

    @property
    def chain(self):
        return self._chain

    @chain.setter
    def chain(self, c):
        self._chain=c

    @property
    def equivalency(self):
        return self._equivalency

    def add_equivalence(self, index):
        self._equivalency.append(index)

    def __str__(self):
        return "Symbol: %s\tPos: %s\tChain: %s\tAdjacent: %s" % (self.symbol, self.pos, self.chain, self.adjacent)

    def to_dict(self):
        return {
        "index": self.index,
        "symbol": self._symbol,
        "adjacent": self.adjacent,
        "pos": self.pos,
        "mass": self._mass,
        "chain": self._chain,
        "equivalency": self._equivalency
        }

    @staticmethod
    def from_dict(in_dict):
        a=Atom(in_dict["symbol"], in_dict["pos"], in_dict["index"], chain=in_dict["chain"])
        a._mass=in_dict["mass"]
        a._adjacent=in_dict["adjacent"]
        a._equivalency=in_dict["equivalency"]
        return a





"""
@author: Devora Witty
"""
import json
import logging
from collections import OrderedDict
from pathlib import Path

import copy
import numpy as np
import os
from openbabel import OBAtomAtomIter, OBConversion, OBMol, OBMolAtomIter, obErrorLog, obError

from csm.input_output.formatters import csm_log as print
from csm.molecule.atom import Atom, GetAtomicSymbol
from csm.molecule.normalizations import normalize_coords, de_normalize_coords, calculate_norm_factor

logger = logging.getLogger("csm")

ob_debug = False
if not ob_debug:
    obErrorLog.SetOutputLevel(obError)


def get_format(format, filename):
    if not format:
        format = filename.split(".")[-1]
    if format.lower() == "csm":
        return "csm"
    conv = OBConversion()
    if not conv.SetInFormat(format):
        format = conv.FormatFromExt(filename)
    if not format:
        raise ValueError("Error discovering format from filename " + filename)
    return format


class Chains(OrderedDict):
    '''
    two sets of keys, string keys and integer keys that match the index of the chain
    the inner dict is composed of integer keys, string keys must be translated to int before sending on to inner dict
    '''

    def __init__(self):
        self._indexes_to_strings = []
        self._strings_to_indexes = {}
        super().__init__()

    def __getitem__(self, key):
        if isinstance(key, int):
            return super().__getitem__(key)
        if isinstance(key, str):
            return super().__getitem__(self._strings_to_indexes[key])
        raise ValueError

    def __setitem__(self, key, value):
        '''
        :param key: expects a string and will force ints to string
        :param value:
        :return:
        '''
        if isinstance(key, int):
            key = str(key)
        self._indexes_to_strings.append(key)
        index = self._indexes_to_strings.index(key)
        self._strings_to_indexes[key] = index
        return super().__setitem__(index, value)

    def __contains__(self, item):
        if isinstance(item, int):
            return item < len(self._indexes_to_strings)
        if isinstance(item, str):
            try:
                return self._strings_to_indexes.__contains__(item)
            except KeyError:
                return False

    def to_array(self):
        zipped_arr = [(key, self.__getitem__(key)) for key in self._indexes_to_strings]
        return zipped_arr

    def from_array(self, arr_of_tuples):
        for key, val in arr_of_tuples:
            self.__setitem__(key, val)

    def index_to_string(self, index):
        return self._indexes_to_strings[index]


class MoleculeMetaData:
    '''
    This class is primarily used to store metadata needed to write results, although format+filecontent+babel_bond
    are sometimes used to recreate a molecule from scratch, see: redo_molecule

    format, filename, and babel_bond are set in _initialize_single_molecule
    file_content is set in read_obm or read_csm
    index is set in read_multiple_molecules, or, revoltingly, in do_commands after calling redo_molecule
    '''

    def __init__(self, file_content=[], format=None, filepath="", babel_bond=False, index=0, initial_title="",
                 initial_comments="", use_filename=True, out_format=None):
        self.file_content = file_content
        self.format = format
        self._out_format = out_format
        self.filepath = filepath
        self.babel_bond = babel_bond
        self.index = index
        self.initial_title = initial_title
        self.initial_comments = initial_comments
        self.use_filename = use_filename

    @property
    def filename(self):
        return os.path.basename(self.filepath)

    @property
    def out_format(self):
        if self._out_format:
            return self._out_format
        return self.format

    @staticmethod
    def from_dict(self, dict):
        m = MoleculeData(**dict)
        return m

    def to_dict(self):
        return vars(self)

    def appellation(self, no_file_format=False, no_leading_zeros=False):
        '''
        the name for the molecule when printing to screen or creating tables.
        if the molecule was read from a folder, it's the filename
        otherwise it's the internal index
        :param no_file_format: remove file ending (eg .xyz)
        :param no_leading_zeros: print the index without leading zeroes
        :return: 
        '''
        if self.use_filename:
            if no_file_format:
                return Path(self.filename).stem
            return self.filename

        mol_index = self.index + 1  # start from 1 instead of 0
        if no_leading_zeros:
            return str(mol_index)

        mol_str = "%04d" % mol_index
        return mol_str


class Molecule:
    """
    Represents a molecule for CSM calculation.
    """

    def __init__(self, atoms=[], to_copy=False):
        """  
        :param atoms: a list of Atoms, default empty
        :param to_copy: boolean, default False, when True none of the molecules fields will be filled
        """
        if not to_copy:
            self._atoms = atoms
            self._norm_factor = 1.0  # is recalculated by molecule._complete_initialization

            self._calculate_center_of_mass()
            self._create_bondset()
            self.create_Q()

            # not included:
            # equivalence class initialization, chain initialization
            self._equivalence_classes = []
            self._chains = Chains()
            self._groups_with_internal_chains = []
            self._chains_with_internal_groups = {}
            self._chain_equivalences = []

            # getting rid of obmol:
            # self._obmol = obmol
            self._deleted_atom_indices = []
            self.has_been_normalized = None

            self.metadata = MoleculeMetaData()

    def __str__(self):
        return self.metadata.appellation()

    def copy(self):
        # deepcopy is used only for atoms,
        # because atoms are the only property changed between runs of Directions that necessitated copying
        # (due to call to denormalize)

        m = Molecule(to_copy=True)
        m._atoms = copy.deepcopy(self.atoms)
        m._norm_factor = self.norm_factor
        m.has_been_normalized = self.has_been_normalized

        m._deleted_atom_indices = self._deleted_atom_indices
        m.metadata = self.metadata

        m._bondset = self.bondset
        m._Q = self.Q
        m._center_of_mass = self._center_of_mass

        m._equivalence_classes = self.equivalence_classes
        m._chains = self._chains
        m._groups_with_internal_chains = self._groups_with_internal_chains
        m._chains_with_internal_groups = self._chains_with_internal_groups
        m._chain_equivalences = self._chain_equivalences

        return m

    def to_dict(self):
        return {
            # critical to include
            "atoms": [atom.to_dict() for atom in self.atoms],

            # trivial to include
            "norm factor": self.norm_factor,
            "center of mass": self.center_of_mass,
            "normalized": self.has_been_normalized,

            # expensive to recalculate
            "equivalence classes": self.equivalence_classes,  # the most expensive part of loading large molecule
            "groups_with_internal_chains": self.groups_with_internal_chains,
            "chains_with_internal_groups": self.chains_with_internal_groups,
            "chain_equivalences": self.chain_equivalences,

            # Classes:
            # obmol: needed for printing:
            "deleted indices": self._deleted_atom_indices,
            "metadata": self.metadata.to_dict(),

            # chains
            "chains": self.chains.to_array()

            # unsure whether worth the bother of including
            # "bondset":self.bondset,
            # "Q":self.Q, #almost definitely not worth the bother
        }

    @staticmethod
    def from_json(raw_json):
        json_dict = json.loads(raw_json)
        return Molecule.from_dict(json_dict)

    @staticmethod
    def from_dict(in_dict):
        atoms = [Atom.from_dict(a) for a in in_dict["atoms"]]
        m = Molecule(atoms)

        m.norm_factor = in_dict["norm factor"]
        m.has_been_normalized = in_dict["normalized"]

        c = Chains()
        c.from_array(in_dict["chains"])
        m._chains = c

        m._deleted_atom_indices = in_dict["deleted indices"]
        m.metadata = MoleculeMetaData.from_dict(in_dict["metadata"])

        m._center_of_mass = in_dict["center of mass"]
        m._equivalence_classes = in_dict["equivalence classes"]
        unfixed_groups = in_dict["groups_with_internal_chains"]
        m._groups_with_internal_chains = []
        for group in unfixed_groups:
            fixed = {int(key): value for key, value in group.items()}
            m._groups_with_internal_chains.append(fixed)
        unfixed_chains = in_dict["chains_with_internal_groups"]
        fixed = {int(key): value for key, value in unfixed_chains.items()}
        m._chains_with_internal_groups = fixed
        m._chain_equivalences = in_dict["chain_equivalences"]

        m.create_Q()
        m._create_bondset()
        return m

    @property
    def center_of_mass(self):
        return self._center_of_mass

    @property
    def Q(self):
        '''
        :return: a numpy array of the molecule's atoms' coordinates, each represented as a numpy array
        '''
        return self._Q

    @property
    def atoms(self):
        '''
        :return: a list of atom objects
        '''
        return self._atoms

    @property
    def bondset(self):
        '''
        :return: a set of tuples, each representing a pair of atoms with a bond between them
        '''
        return self._bondset

    @property
    def equivalence_classes(self):
        '''
        :return: a list of the molecule's equivalence classes (themselves represented as lists), sorted in order of length
        '''
        self._equivalence_classes.sort(key=len)
        return self._equivalence_classes

    @property
    def norm_factor(self):
        # Normalization factor. Defaults to 1.0 if normalize wasn't called
        return self._norm_factor

    @norm_factor.setter
    def norm_factor(self, nf):
        self._norm_factor = nf

    @property
    def chains(self):
        '''
        :return: a dictionary of the molecule's chains, with keys being the name of the chains
        '''
        return self._chains

    @property
    def groups_with_internal_chains(self):
        return self._groups_with_internal_chains

    @property
    def chains_with_internal_groups(self):
        return self._chains_with_internal_groups

    @property
    def chain_equivalences(self):
        return self._chain_equivalences

    def __len__(self):
        return len(self._atoms)

    def _create_bondset(self):
        self._bondset = set()
        for i in range(len(self._atoms)):
            for match in self._atoms[i].adjacent:
                self._bondset.add((i, match))

    def _initialize_chains(self, use_chains):
        """
        1. Create chains, and overwrite atom.chain with index of chain rather than string
        2. within each equivalence class, labels by chain
            self._groups_with_internal_chains=[array of equivalence classes:
                            {dictionary of chains:
                                [array of indexes belonging to that chain in that equivalence class]
                            }
                        ]
            self._chains_with_internal_groups= {dicitonary of chains:
        3. Calculates chain equivalencies. (e.g, chains A and B are equivalent, chains C and D are equivalent)
        """

        # create chains, and overwrite atom.chain with index of chain rather than string
        chains = Chains()
        for i, atom in enumerate(self.atoms):
            chain = atom.chain
            if not use_chains:
                chain = 'Simulated chain'
            if chain not in chains:
                chains[chain] = []
            chains[chain].append(i)
            atom.chain = chains._strings_to_indexes[chain]
        self._chains = chains

        # within each equivalence class, labels by chain,
        # within each chain, group by equivalence class
        groups_with_internal_chains = []
        num_equiv = len(self.equivalence_classes)
        chains_with_internal_groups = {chain: [None] * num_equiv for chain in self.chains}

        for group_index, group in enumerate(self.equivalence_classes):
            chaingroup = {}
            for atom_index in group:
                chain_index = self.atoms[atom_index].chain

                # add index to array in chaingroup dict
                try:
                    chaingroup[chain_index].append(atom_index)
                except KeyError:
                    chaingroup[chain_index] = [atom_index]
                # add index to classed_chains
                try:
                    chains_with_internal_groups[chain_index][group_index].append(atom_index)
                except AttributeError:
                    chains_with_internal_groups[chain_index][group_index] = [atom_index]
            # add the completed chaingroup dict to group_chains arrat
            groups_with_internal_chains.append(chaingroup)

        self._groups_with_internal_chains = groups_with_internal_chains
        self._chains_with_internal_groups = chains_with_internal_groups

        # Calculates chain equivalencies. (e.g, chains A and B are equivalent, chains C and D are equivalent)
        # we define chains as equivalent if the number of atoms they have in each equivalence class are the same
        chain_equivalences = []
        marked = [False] * len(self.chains)

        for chain_index, chainkey in enumerate(self.chains):
            if marked[chain_index]:
                continue
            equiv = []
            for chain2_index, chainkey2 in enumerate(self.chains):
                # start with a simple length check to spare checking equivalence classes if chains arent same length to begin with
                # if len(self.chains[chainkey])!=len(self.chains[chainkey2]):
                #    continue
                same_lengths = True
                for group in groups_with_internal_chains:
                    if not same_lengths:
                        break
                    try:
                        length = len(group[chain_index])
                    except KeyError:  # that chain is notn in this equivalence group
                        continue
                    try:
                        if length != len(group[chain2_index]):
                            same_lengths = False
                    except KeyError:
                        same_lengths = False
                if same_lengths:
                    equiv.append(chain2_index)
                    marked[chain2_index] = True
            chain_equivalences.append(equiv)
        self._chain_equivalences = chain_equivalences

    def _calculate_equivalency(self):
        """
        Preprocess a molecule based on the arguments passed to CSM
        """

        def is_similar(atoms_group_num, a, b):
            found = True
            mark = set()

            valency_a = len(self._atoms[a].adjacent)
            valency_b = len(self._atoms[b].adjacent)

            # for each of i's neighbours
            for i in range(valency_a):
                found = False

                for j in range(valency_b):
                    if j in mark:
                        continue

                    if atoms_group_num[self._atoms[a].adjacent[i]] == atoms_group_num[self._atoms[b].adjacent[j]]:
                        # the i-th neighbour of 'a' belongs to the same group as the j-th neighbour of 'b'
                        found = True
                        mark.add(j)
                        break

                if not found:
                    break

            return found

        group_num = 0
        groups = []
        atoms_size = len(self._atoms)
        marked = set()

        atoms_group_num = {}

        # break into initial groups by symbol and valency
        for i in range(atoms_size):
            if i in marked:
                continue

            groups.append([])

            for j in range(atoms_size):
                if j in marked \
                        or len(self._atoms[i].adjacent) != len(self._atoms[j].adjacent) \
                        or self._atoms[i].symbol != self._atoms[j].symbol:
                    continue

                groups[group_num].append(j)
                atoms_group_num[j] = group_num
                marked.add(j)

            group_num += 1

        # logger.debug("initial number of groups:" + str(group_num))
        # iteratively refine the breakdown into groups
        # break into subgroups at an infinite depth - as long as there's something to break, it is broken

        divided_group = True
        num_iters = 0

        while divided_group:
            num_iters += 1
            divided_group = False

            new_groups = []

            for i, group in enumerate(groups):
                first_elem = group[0]
                group_size = len(group)

                sub_group = []

                # for each item in the group (except the first) check if it can be split or not
                for j in range(1, group_size):
                    if not is_similar(atoms_group_num, group[j], first_elem):
                        # add elem to new subGroup
                        sub_group.append(group[j])
                        group[j] = -1

                if len(sub_group) > 0:
                    divided_group = True
                    new_groups.append(sub_group)
                    for el in sub_group:
                        atoms_group_num[el] = group_num
                    group_num += 1
                    # remove elements of sub_group from group
                    groups[i] = [el for el in group if el != -1]

            groups.extend(new_groups)

        # logger.debug("Broken into %d groups with %d iterations." % (group_num, num_iters))

        self._equivalence_classes = groups
        for group in groups:
            for atom_index in group:
                for equiv_index in group:
                    self._atoms[atom_index].add_equivalence(equiv_index)

    def strip_atoms(self, remove_hy=False, select_atoms=[]):
        """
            Creates a new Molecule from m by removing atoms who's symbol is in the remove list
            :param csm_args:
            :param removeList: atomic symbols to remove
        """

        removed_atoms = []
        fixed_indexes = [i for i in range(len(self))]

        for i in range(len(self._atoms)):
            if remove_hy:
                if self._atoms[i].symbol == "H":
                    removed_atoms.append(i)
                    fixed_indexes[i] = None
                else:
                    # however many atoms have been removed up to this index is the amount its index needs adjusting by
                    fixed_indexes[i] -= len(removed_atoms)

            if select_atoms:
                if i not in select_atoms:
                    removed_atoms.append(i)
                    fixed_indexes[i] = None
                else:
                    # however many atoms have been removed up to this index is the amount its index needs adjusting by
                    fixed_indexes[i] -= len(removed_atoms)

        # adjust the connectivity indices before we do any popping whatsoever
        for i, atom in enumerate(self._atoms):
            adjacent_new = []
            for adjacent in atom.adjacent:
                if fixed_indexes[adjacent] is not None:
                    adjacent_new.append(fixed_indexes[adjacent])
            atom.adjacent = adjacent_new

        for to_remove in reversed(removed_atoms):  # reversed order because popping changes indexes after
            self._atoms.pop(to_remove)
            self._deleted_atom_indices.append(to_remove)
            # if remove_hy: #this is meant to affect print at end
            # self._obmol.DeleteAtom(self._obmol.GetAtom(to_remove + 1))

        self._create_bondset()
        # logger.debug(len(removed_atoms), "atoms removed")

    def _calculate_center_of_mass(self):
        coords = [atom.pos for atom in self._atoms]
        masses = [atom.mass for atom in self._atoms]
        x_avg = y_avg = z_avg = 0.0

        size = len(masses)

        mass_sum = 0
        for i in range(size):
            x_avg += coords[i][0] * masses[i]
            y_avg += coords[i][1] * masses[i]
            z_avg += coords[i][2] * masses[i]
            mass_sum += masses[i]
        x_avg /= mass_sum
        y_avg /= mass_sum
        z_avg /= mass_sum

        self._center_of_mass = [x_avg, y_avg, z_avg]

    def normalize(self):
        """
        Normalize the molecule
        """
        coords = [atom.pos for atom in self._atoms]
        self._calculate_center_of_mass()
        self._norm_factor = calculate_norm_factor(coords, self.center_of_mass)
        norm_coords = normalize_coords(coords, self.center_of_mass, self.norm_factor)
        size = len(self._atoms)
        for i in range(size):
            self._atoms[i].pos = norm_coords[i]
        self.create_Q()
        self.has_been_normalized = True

    def de_normalize(self):
        coords = [atom.pos for atom in self._atoms]
        denorm_coords = de_normalize_coords(coords, self.norm_factor)

        size = len(self._atoms)
        for i in range(size):
            self._atoms[i].pos = denorm_coords[i]
        self.create_Q()
        self.has_been_normalized = False

    def create_Q(self):
        self._Q = np.array([np.array(atom.pos) for atom in self.atoms])

    def print_equivalence_class_summary(self, display_chains=True):
        """
        Displays information about equivalence classes and chains
        """
        lengths = {}
        for group in self._equivalence_classes:
            try:
                lengths[len(group)] += 1
            except KeyError:
                lengths[len(group)] = 1
        for key in lengths:
            print("%d group%s of length %d" % (lengths[key], 's' if lengths[key] and lengths[key] > 1 else '', key))

        if True:
                for chain in self.chains:
                    print("Chain %s of length %d" % (self.chains._indexes_to_strings[chain], len(self.chains[chain])))
                print("%d equivalence class%s of chains" % (len(self.chain_equivalences), 'es' if lengths[key] else ''))
                for chaingroup in self.chain_equivalences:
                    chainstring = "Group of length " + str(len(chaingroup)) + ":"
                    for index in chaingroup:
                        chainstring += " "
                        chainstring += str(self.chains._indexes_to_strings[index])
                    print(str(chainstring))
            # else:
            #    print("Molecule has no chains")

    def _complete_initialization(self, use_chains, remove_hy, select_atoms=[]):
        """
        Finish creating the molecule after reading the raw data
        """
        self.strip_atoms(remove_hy, select_atoms)
        self._calculate_equivalency()
        self._initialize_chains(use_chains)
        self.normalize()

    @staticmethod
    def xyz_string(atoms, positions=None, header=""):
        '''
        :param atoms: an array of atoms
        :param positions: optional, an array of positions to overwrite positions in the array of atoms
        :return:
        '''
        coords = str(len(atoms))
        coords += "\n" + header + "\n"
        for i, atom in enumerate(atoms):
            coords += str(atom.symbol)
            for index, coor in enumerate(atom.pos):
                coords += "\t"
                if positions is not None:
                    coords += str(positions[i][index])
                else:
                    coords += str(coor)
            coords += "\n"
        return coords


class MoleculeFactory:
    @staticmethod
    def dummy_molecule_from_size(size, groups):
        '''
        The dummy molecule is a fake molecule used in the approx algorithm to generate chain permutations
        :param size: number of "atoms" in the fake molecule
        :param groups: the equivalence classes of those "atoms"
        :return:
        '''
        atoms = []

        for i in range(size):
            atom = Atom("C", (0, 0, 0), i, False)
            atoms.append(atom)
        mol = Molecule(atoms)
        mol._equivalence_classes = groups
        return mol

    @staticmethod
    def dummy_molecule_from_coords(coords, groups=None):
        atoms = []

        if groups is None:
            groups = [[i for i in range(len(coords))]]

        for fake_atom, group in enumerate(groups):
            for atom_index in group:
                coord = coords[atom_index]
                atom = Atom(str(fake_atom), coord, atom_index, False)
                atoms.append(atom)

        mol = Molecule(atoms)
        mol._calculate_equivalency()
        mol.normalize()
        return mol


class PDBLine:
    '''
    Currently only supports ATOM, HETATM, and CONECT records-- if any others are needed in the future they can be added then
    '''

    def __init__(self, pdb_line):
        self.pdb_line = pdb_line
        self._record_name = pdb_line[0:6]
        self.record_name = self._record_name.strip()

        self._atom_serial_number = pdb_line[6:11]  # used in read pdb connectivity
        if self.record_name in ["ATOM", "HETATM"]:
            self._init_atom_record(pdb_line)

        if self.record_name == "CONECT":
            self._init_conect_record(pdb_line)

    def _init_atom_record(self, pdb_line):
        self.full_atom_name = pdb_line[12:16]
        self.alternate_location_indicator = pdb_line[16]
        self.residue_name = pdb_line[17:20]

        # handle chains
        # ATOMxxxxxx1xxNxxxLYSxA
        # HETATMxxxx3xxHxxxHOHxxxxx1
        if self.record_name == 'ATOM':
            chain_designation = self.pdb_line[21]
        if self.record_name == 'HETATM':
            if self.pdb_line[21] != " ":
                chain_designation = self.pdb_line[21]
            else:
                chain_designation = self.pdb_line[25]
        self.chain_id = chain_designation

        self.sequence_number = pdb_line[22:26],  # used in use sequence
        self.iCode = pdb_line[26]
        self.x = pdb_line[30:38]
        self.y = pdb_line[38:46]
        self.z = pdb_line[46:54]
        self.occ = pdb_line[54:60]
        self.temp_factor = pdb_line[60:66]
        self.element = pdb_line[76:78]
        self.charge = pdb_line[78:80]

        self._parse_atom_name(pdb_line)

    def _parse_atom_name(self, pdb_line):
        # https://www.cgl.ucsf.edu/chimera/docs/UsersGuide/tutorials/pdbintro.html
        # Columns 13-16 - the atom name. Columns 13-14 contain the atom symbol, right aligned (except when the atom is for H).
        # Column 15 contains the remoteness indicator that represents the distance of the atom from the backbone of the protein. Greek letter remoteness codes are transliterated as follows:
        # alpha=A, beta=B, gamma = G, delta = D, epsilon=E, zeta= Z, eta=H].
        # Column 16 – is the branch designator
        # Column 17 - Alternate location.
        # *Atom names start with element symbols right-justified in columns 13-14 as permitted by the length of the name.
        # For example, the symbol FE for iron appears in columns 13-14, whereas the symbol C for carbon appears in column 14 (see Misaligned Atom Names).
        # If an atom name has four characters, however, it must start in column 13 even if the element symbol is a single character (for example, see Hydrogen Atoms).
        # Hydrogen atom records follow the records of all other atoms of a particular residue.
        # A hydrogen atom name starts with H. The next part of the name is based on the name of the connected nonhydrogen atom.
        # For example, in amino acid residues, H is followed by the remoteness indicator (if any) of the connected atom, followed by the branch indicator (if any) of the connected atom;
        # if more than one hydrogen is connected to the same atom, an additional digit is appended so that each hydrogen atom will have a unique name.
        # Hydrogen atoms in standard nucleotides and amino acids (other than the rarely seen HXT) are named according to the IUPAC recommendations (Pure Appl Chem 70:117 (1998) [abstract] [PDF]).
        # Names of hydrogen atoms in HETATM residues are determined in a similar fashion.
        # If the name of a hydrogen has four characters, it is left-justified starting in column 13; if it has fewer than four characters, it is left-justified starting in column 14.

        atom_name = pdb_line[12:16]
        atom_name = atom_name.strip()
        if atom_name[0] == "H":
            self.atom_symbol = "H"
            self.remoteness = ""
            self.branch_designation = ""
            self.duplicate_index = ""
            self.alternate_location = pdb_line[16]
            try:
                self.remoteness = self.atom_symbol[1]
                self.branch_designation = self.atom_symbol[2]
                self.duplicate_index = self.atom_symbol[3]
            except:  # TODO: check that what information is available gets recorded
                pass  # because it may be attempting to access information that doesn't exist so it's ok...
        else:
            self.atom_symbol = pdb_line[12:14].strip()
            self.remoteness = pdb_line[14]
            self.branch_designation = pdb_line[15]
            self.alternate_location = pdb_line[16]

        if not self.atom_symbol:
            raise ValueError(
                "PDB missing atom symbol information in atom name (columns 13-16). Please check that your pdb is formatted correctly")

    def _init_conect_record(self, pdb_line):
        adjacent_atoms = []
        stripped_pdb_line = pdb_line.strip()
        for i in range(11, len(stripped_pdb_line), 5):
            adjacent_atom_index = int(stripped_pdb_line[i:i + 5])
            adjacent_atoms.append(adjacent_atom_index)
        self.adjacent_atoms = adjacent_atoms

    @property
    def atom_serial_number(self):
        try:
            return int(self._atom_serial_number)
        except (ValueError, IndexError):
            return None

    @staticmethod
    def _pdb_line_to_dict(pdb_line):
        return PDBLine(pdb_line)


class MoleculeReader:
    """
    A static class that creates instances of Molecule from files or strings.
    """

    @staticmethod
    def _remove_multi_bonds(bonds):
        """
        Receives a list, returns a sorted list with no duplicates.
        :param bonds: a list
        :return: a sorted list with no duplicates
        """
        l = list(set(bonds))
        l.sort()
        return l

    @staticmethod
    def from_string(string, format, initialize=True,
                    use_chains=False, babel_bond=False,
                    remove_hy=False, ignore_sym=False, use_mass=False):
        """
        create a Molecule from a string
        :param string: the string to create the molecule from
        :param format: the format of the string (any BabelBond supported format, eg "mol", "xyz")
        :param initialize: boolean, default True, when True equivalence classes and chains are calculated for the molecule
        :param use_chains: boolean, default False, when True chains are read from the string 
        :param babel_bond: boolean, default False, when True OpenBabel will attempt to guess connectivity information
        :param remove_hy: boolean, default False, when True hydrogen atoms will be removed
        :param ignore_sym: boolean, default False, when True all atom symbols will be "X" instead of elements 
                            (this affects equivalence class calculation)
        :param use_mass: boolean, default False, when True atomic mass will be used, when False all atoms have a mass of 1
        :return: an instance of class Molecule
        """
        # note: useMass is used when creating molecule, even though it is actually about creating the normalization
        # step one: get the molecule object
        obm = MoleculeReader._obm_from_strings([string], format, babel_bond)
        mol = MoleculeReader.mol_from_obm(obm, format, ignore_sym=ignore_sym, use_mass=use_mass)
        if initialize:
            mol._complete_initialization(use_chains, remove_hy)
        return mol

    @staticmethod
    def from_file(in_file_name, in_format=None, initialize=True,
                  use_chains=False, babel_bond=False,
                  remove_hy=False, ignore_sym=False, use_mass=False,
                  read_fragments=False, use_sequence=False,
                  keep_structure=False, select_atoms=[], conn_file=None,
                  out_format=None,
                  *args, **kwargs):
        """
        :param in_file_name: the name of the file to read the molecule from
        :param in_format: the format of the string (any BabelBond supported format, eg "mol", "xyz")
        :param initialize: boolean, default True, when True equivalence classes and chains are calculated for the molecule
        :param use_chains: boolean, default False, when True chains are read from the string
        :param babel_bond: boolean, default False, when True OpenBabel will attempt to guess connectivity information
        :param remove_hy: boolean, default False, when True hydrogen atoms will be removed
        :param ignore_sym: boolean, default False, when True all atom symbols will be "X" instead of elements
                            (this affects equivalence class calculation)
        :param use_mass: boolean, default False, when True atomic mass will be used, when False all atoms have a mass of 1
        :param read_fragments: boolean, default False, when True, multiple molecules in one file will be treated
                                as "chains" of one composite molecule
        :param use_sequence: boolean, default False, for pdb files only-- uses pdb sequence information to establish
                                equivalence (overwrites default equivalence)
        :param keep_structure: boolean, default False. If keep-structure is specified and molecule has no bond information,
                                the peogram will raise an error and exit
        :return: an instance of class Molecule
        """
        # suppress warnings from openbabel

        format = get_format(in_format, in_file_name)
        if True:
            obm = MoleculeReader._obm_from_file(in_file_name, format, babel_bond)
            mol = MoleculeReader.mol_from_obm(obm, format, ignore_sym=ignore_sym, use_mass=use_mass,
                                              read_fragments=read_fragments)
        return MoleculeReader._process_single_molecule(mol, in_file_name, format, initialize,
                                                      use_chains, babel_bond,
                                                      remove_hy, ignore_sym, use_mass,
                                                      read_fragments, use_sequence,
                                                      keep_structure, select_atoms, conn_file,
                                                      out_format)

    @staticmethod
    def _process_single_molecule(mol, in_file_name, format, initialize=True,
                                 use_chains=False, babel_bond=False,
                                 remove_hy=False, ignore_sym=False, use_mass=False,
                                 read_fragments=False, use_sequence=False,
                                 keep_structure=False, select_atoms=[], conn_file=None,
                                 out_format=None, **kwargs):

        mol.metadata.format=format
        mol.metadata.babel_bond=babel_bond
        mol.metadata.filepath=in_file_name

        if use_sequence:
            if format.lower() == 'pdb':
                mol = MoleculeReader._create_pdb_with_sequence (mol, in_file_name, use_chains=use_chains,
                                                           babel_bond=babel_bond,
                                                           read_fragments=read_fragments, remove_hy=remove_hy,
                                                           ignore_sym=ignore_sym, use_mass=use_mass)
                # we initialize mol from within pdb_with_sequence because otherwise equivalnce classes would be overwritten
                return mol

        if format == "pdb":
            mol = MoleculeReader._read_pdb_connectivity_and_chains(in_file_name, mol, read_fragments, babel_bond)
        if conn_file and format=="xyz":
            MoleculeReader.read_xyz_connectivity(mol, conn_file)
        if not mol.bondset:
            if keep_structure:
                raise ValueError(
                    "User input --keep-structure but input molecule has no bonds. Did you forget --babel-bond?")
            else:
                print("Warning: Input molecule has no bond information")

        if initialize:
            mol._complete_initialization(use_chains, remove_hy, select_atoms)
            if len(mol.chains) < 2:
                raise ValueError("The molecule you have provided has less than 2 chains. "
                                 "Protein-csm only works with molecules with 2 or more chains.")
        return mol

    @staticmethod
    def multiple_from_file(in_file_name, in_format=None, initialize=True,
                           use_chains=False, babel_bond=False,
                           remove_hy=False, ignore_sym=False, use_mass=False,
                           read_fragments=False, use_sequence=False,
                           keep_structure=False, select_atoms=[], conn_file=None,
                           out_format=None,
                           *args, **kwargs):
        mols = []
        format = get_format(in_format, in_file_name)

        if format == "csm":
            mol = MoleculeReader._read_csm_file(in_file_name, ignore_sym, use_mass)
            mols.append(mol)

        else:
            obms = MoleculeReader._obm_from_file(in_file_name, format, babel_bond)
            if read_fragments:
                mol = MoleculeReader.mol_from_obm(obms, format, babel_bond=babel_bond, ignore_sym=ignore_sym,
                                                  use_mass=use_mass, read_fragments=read_fragments)
                mols.append(mol)

            else:
                for obm in obms:
                    mol = MoleculeReader.mol_from_obm([obm], format, babel_bond=babel_bond, ignore_sym=ignore_sym,
                                                      use_mass=use_mass)
                    mols.append(mol)

        processed_mols = []
        use_filename = True
        if len(mols) > 1:
            use_filename = False
        for index, mol in enumerate(mols):
            p_mol = MoleculeReader._process_single_molecule(mol, in_file_name, format, initialize,
                                                            use_chains, babel_bond,
                                                            remove_hy, ignore_sym, use_mass,
                                                            read_fragments, use_sequence,
                                                            keep_structure, select_atoms, conn_file,
                                                            out_format)
            p_mol.metadata.index = index
            p_mol.metadata.use_filename = use_filename
            processed_mols.append(p_mol)
        return processed_mols

    @staticmethod
    def _obm_from_strings(strings, format, babel_bond=False):
        '''
        :param strings: an array of molecule strings
        :param format:
        :param babel_bond:
        :return:
        '''
        conv = OBConversion()
        obmol = OBMol()
        if not conv.SetInFormat(format):
            raise ValueError("Error setting openbabel format to" + format)
        if not babel_bond:
            conv.SetOptions("b", conv.INOPTIONS)
        obmols = []
        for string in strings:
            conv.ReadString(obmol, string)
            obmols.append(obmol)
            obmol = OBMol()
        return obmols

    @staticmethod
    def _obm_from_file(filename, format=None, babel_bond=False):
        """
        :param filename: name of file to open
        :param format: molecule format of file (eg xyz, pdb)
        :param babelBond:
        :return:
        """
        obmol = OBMol()
        conv = OBConversion()
        if not format:
            format = get_format(format, filename)
        if not conv.SetInFormat(format):
            raise ValueError("Error setting openbabel format to" + format)
        if format == "txt":
            raise ValueError("CSM does not support .txt format openbabel conversions")
        if not babel_bond:
            conv.SetOptions("b", conv.INOPTIONS)
        notatend = conv.ReadFile(obmol, filename)
        if not notatend:
            raise ValueError("Error reading file " + filename + " using OpenBabel, with format:", format)
        obmols = []
        while notatend:
            obmols.append(obmol)
            obmol = OBMol()
            notatend = conv.Read(obmol)
        return obmols

    @staticmethod
    def mol_from_obm(obmols, format, babel_bond=False, ignore_sym=False, use_mass=False, read_fragments=False,
                     **kwargs):
        """
        :param obmol: OBmol molecule
        :param args_dict: dictionary of processed command line arguments
        :return: A list of Atoms and a list of chains
        """
        if not read_fragments:
            obmols = obmols[:1]
        atoms = []
        mol_contents = []
        for obmol_id, obmol in enumerate(obmols):
            title_contents = obmol.GetTitle()
            mol_contents.append(mol_string_from_obm(obmol, format))
            for i, obatom in enumerate(OBMolAtomIter(obmol)):
                position = (obatom.GetX(), obatom.GetY(), obatom.GetZ())
                if ignore_sym:
                    symbol = "XX"
                else:
                    # get symbol by atomic number
                    symbol = GetAtomicSymbol(obatom.GetAtomicNum())
                try:
                    if read_fragments:
                        chain = str(obmol_id)
                    else:
                        chain = obatom.GetResidue().GetChain().strip()
                    atom = Atom(symbol, position, i, use_mass, chain)
                except (AttributeError, KeyError):
                    # molecule doesn't have chains, stuck with empty chains
                    atom = Atom(symbol, position, i, use_mass)
                adjacent = []
                for neighbour_atom in OBAtomAtomIter(obatom):
                    adjacent.append(neighbour_atom.GetIdx() - 1)
                atom.adjacent = MoleculeReader._remove_multi_bonds(adjacent)
                atoms.append(atom)

        mol = Molecule(atoms)
        mol.metadata.file_content = mol_contents
        mol.metadata._title = title_contents
        return mol


    @staticmethod
    def _read_pdb_connectivity_and_chains(filename, mol, read_fragments, babel_bond):
        with open(filename, 'r') as file:
            # Count ATOM and HETATM lines, mapping them to our ATOM numbers.
            # In most PDBs ATOM 1 is our atom 0, and ATOM n is our n-1. However, in some cases
            # there are TER lines in the PDB, ATOMs after the TER are found at n-(TERCOUNT+1) in our list.
            # Some pdb files have multiple models even though we're only reading the first
            atom_map = {}
            cur_atom = 0
            # chains = Chains()

            for line in file:
                pdb_dict = PDBLine._pdb_line_to_dict(line)
                index = pdb_dict.atom_serial_number

                record_name = pdb_dict.record_name
                if record_name in ['ATOM', 'HETATM'] and cur_atom < len(mol):
                    if record_name == 'HETATM':
                        breakpt = 1
                    atom_map[index] = cur_atom

                    chain_designation = pdb_dict.chain_id
                    if chain_designation != " ":
                        mol._atoms[cur_atom]._chain = str(chain_designation)

                    cur_atom += 1

                if not babel_bond:
                    if record_name == "CONECT":
                        # CONECT records are described here: http://www.bmsc.washington.edu/CrystaLinks/man/pdb/part_69.html
                        # After CONECT appears a series of 5 character atom numbers. There are no separating spaces in case
                        # the atom numbers have five digits, so we need to split the atom numbers differently

                        try:
                            fake_atom_index = pdb_dict.atom_serial_number
                            try:
                                atom_index = atom_map[fake_atom_index]
                            except KeyError:
                                if fake_atom_index > len(mol):
                                    continue
                            atom = mol._atoms[atom_index]

                            # add adjacency
                            adjacent = []
                            for adjacent_atom_index in pdb_dict.adjacent_atoms:
                                adjacent.append(atom_map[adjacent_atom_index])
                            atom.adjacent = MoleculeReader._remove_multi_bonds(adjacent)
                        except Exception as e:
                            raise ValueError("There was a problem reading connectivity from the pdb file." + str(e))
                    mol._create_bondset()  # refresh the bondset
        return mol

    @staticmethod
    def _create_pdb_with_sequence(mol, in_file_name, initialize=True,
                                  use_chains=False, babel_bond=False, read_fragments=False,
                                  ignore_hy=False, remove_hy=False, ignore_sym=False, use_mass=False):
        def read_atom(line, likeness_dict, cur_atom):
            pdb_dict = PDBLine._pdb_line_to_dict(line)
            atom_type = pdb_dict.atom_symbol
            remoteness = pdb_dict.remoteness
            serial_number = pdb_dict.sequence_number
            key = tuple([atom_type, remoteness, serial_number])
            if key not in likeness_dict:
                likeness_dict[key] = [cur_atom]
            else:
                likeness_dict[key].append(cur_atom)

        def set_equivalence_classes(mol, likeness_dict):
            groups = []
            for key in likeness_dict:
                groups.append(likeness_dict[key])

            mol._equivalence_classes = groups
            try:
                for group in groups:
                    for atom_index in group:
                        for equiv_index in group:
                            mol._atoms[atom_index].add_equivalence(equiv_index)
            except Exception as e:  # TODO: comment why this except is here (I don't actually remember)
                print(e)

        mol = MoleculeReader._read_pdb_connectivity_and_chains(in_file_name, mol, read_fragments, babel_bond)
        if remove_hy or ignore_hy:
            mol.strip_atoms(remove_hy, ignore_hy)
        likeness_dict = {}
        cur_atom = 0

        with open(in_file_name, 'r') as file:
            for line in file:
                pdb_dict = PDBLine(line)
                if pdb_dict.record_name in ['ATOM', 'HETATM'] and cur_atom < len(mol):
                    if remove_hy or ignore_hy:
                        if pdb_dict.atom_symbol == "H":
                            continue
                    read_atom(line, likeness_dict, cur_atom)
                    cur_atom += 1

        mol._create_bondset()
        mol.create_Q()

        # we have our own equivalence class function and hence cannot call the main initialization
        set_equivalence_classes(mol, likeness_dict)

        mol._initialize_chains(use_chains)
        mol.normalize()

        return mol

def mol_string_from_obm(obmol, format):
    obconversion = OBConversion()
    formatok = obconversion.SetOutFormat(format)
    if not formatok:
        raise ValueError("%s is not a recognised Open Babel format" %
                         format)
    return obconversion.WriteString(obmol)



"""
Parse the CSM command line arguments.
"""
from argparse import ArgumentParser
import logging

from csm.calculations.data_classes import Operation

logger = logging.getLogger(__name__)
import sys


__author__ = 'zmbq'

class OurParser(ArgumentParser):
    def error(self, message):
        print("Error: %s" % message, file=sys.stderr)
        print("Run proteincsm --help for help", file=sys.stderr)
        sys.exit(2)


def _create_parser():
    parser = OurParser(usage="\nproteincsm type input_molecule output_file [additional arguments]", allow_abbrev=False)

    parser.add_argument('type',
                        # choices=c_symmetries + s_symmetries + ['cs', 'ci', 'ch'],
                        help='The type of operation: cs, ci, ch, cN, sN')
    parser.add_argument('input', help='Input molecule file')
    parser.add_argument('output', default='output.txt', help='Output file')

    # Optional arguments

    # types of calculations:
    calculation_type = parser.add_argument_group('Calculation Type (default is approx with hungarian algorithm)')
    _calculation_type=calculation_type.add_mutually_exclusive_group()
    _calculation_type.add_argument('--trivial', action='store_const', const='trivial', default='approx', dest='calc_type',
                                  help='ProteinCSM of identity perm, or, if chains, CSM of chain permutation with no atom permutation')
    _calculation_type.add_argument('--greedy', action='store_const', const='greedy', default='many-chains', dest='approx_algorithm',
                             help='Use the old greedy approx algorithm (no hungarian)-- not relevant for trivial')


    parser.add_argument('--timeout', default=300,
                        help="Specify a timeout for the calculation in seconds. Default is 5 minutes (300)", type=int)
    parser.add_argument('--sn-max', type=int, default=8, help='The maximal sn to try, relevant only for chirality')

    # general input/calculation arguments:
    input_type = parser.add_argument_group('Input Arguments')
    input_type.add_argument('--remove-hy', action='store_true', default=False,
                            help='Remove Hydrogen atoms, rebuild molecule without them, and compute')
    input_type.add_argument('--babel-bond', action='store_true', default=False, help='Let OpenBabel compute bonding')
    #input_type.add_argument('--use-sequence', action='store_true', default=False,
    #                        help='create equivalence class for pdb file using sequence information. Can\'t be used with --use-chains')
    #input_type.add_argument('--use-chains', action='store_true', default=False,
    #                        help='When a molecule has chains, use them (affects trivial, approx)')

    return parser

def _process_arguments(parse_res):
    """
    Divides the parsed arguments (from argparse) into three dictionaries
    Args:
        parse_res: Result of argparse

    Returns:
        dictionary_args - input arguments, calculation arguments and output arguments

    """

    dictionary_args = dict(vars(parse_res))

    # the first three positional arguments

    op = Operation(parse_res.type, parse_res.sn_max)

    dictionary_args['operation'] = op
    dictionary_args['op_type'] = op.type
    dictionary_args['op_order'] = op.order
    dictionary_args['op_name'] = op.name

    dictionary_args['in_file_name'] = parse_res.input
    dictionary_args['out_file_name'] = parse_res.output

    dictionary_args["use_sequence"] = True
    dictionary_args["use_chains"] = True

    return dictionary_args


def get_split_arguments(args):
    """
    :param args:
    :return:
    """
    parser = _create_parser()
    parsed_args=parser.parse_args(args)
    #parsed_args, leftovers = parser.parse_known_args(args)
    dictionary_args = _process_arguments(parsed_args)
    return dictionary_args


if __name__ == '__main__':
    args=sys.argv[1:]
    get_split_arguments(args)



"""
Functions for finding the symmetry directions prior to running approx
"""
import math
import numpy as np
from csm.fast import external_get_eigens as cppeigen
from csm.input_output.formatters import csm_log as print
from csm.calculations.constants import MINDOUBLE

MIN_GROUPS_FOR_OUTLIERS = 10


class DirectionChooser:
    '''
    Any direction chooser simply needs a property dirs
    '''
    @property
    def dirs(self):
        return self._dirs

class ClassicDirectionChooser(DirectionChooser):
    '''
    This class implements the classic direction chooser algorithm described in the paper [citation needed].
    '''
    def __init__(self, molecule, op_type, op_order, use_best_dir=False, get_orthogonal=True, detect_outliers=False, *args, **kwargs):
        self._op_type = op_type
        self._op_order = op_order
        self._dirs = self._choose_initial_directions(molecule, use_best_dir, get_orthogonal, detect_outliers,
                                                     op_type)


    def _choose_initial_directions(self, molecule, use_best_dir, get_orthogonal, detect_outliers,
                                           op_type):
        # if inversion:
        # not necessary to calculate dir, use geometrical center of structure
        if self._op_type == 'CI' or (self._op_type == 'SN' and self._op_order == 2):
            initial_directions = [[1.0, 0.0, 0.0]]

        else:
            initial_directions = self.find_symmetry_directions(molecule, use_best_dir, get_orthogonal, detect_outliers,
                                                          op_type)
        return initial_directions
    def find_symmetry_directions(self, molecule, use_best_dir, get_orthogonal, detect_outliers, op_type):
        # get average position of each equivalence group:
        group_averages = []
        for group in molecule.equivalence_classes:
            sum = [0, 0, 0]
            for index in group:
                sum += molecule.Q[index]
            average = sum / len(group)
            group_averages.append(average)
        group_averages = np.array(group_averages)

        dirs = self.dir_fit(group_averages, use_best_dir)

        if detect_outliers and len(molecule.equivalence_classes) > MIN_GROUPS_FOR_OUTLIERS:
            dirs = self.dirs_without_outliers(dirs, group_averages, op_type, use_best_dir)
        if get_orthogonal:
            dirs = self.dirs_orthogonal(dirs)
        return dirs

    def normalize_dir(self,dir):
        dir = np.array(dir)  # lists also get sent to this function, but can't be /='d
        norm = math.sqrt(math.fabs(dir[0] * dir[0] + dir[1] * dir[1] + dir[2] * dir[2]))
        dir /= norm
        return dir

    def dir_fit(self, positions, best_dir):
        # get vector of the average position
        sum = np.einsum('ij->j', positions)
        average = sum / len(positions)

        mat = np.zeros((3, 3), dtype=np.float64, order="c")
        for pos in positions:
            for j in range(3):
                for k in range(3):
                    mat[j][k] += (pos[j] - average[j]) * (pos[k] - average[k])

        # computer eigenvalues and eigenvectors
        # lambdas - list of 3 eigenvalues of matrix (function demands them, but we won't be using them here)
        # dirs- list of 3 eigenvectors of matrix
        dirs = np.zeros((3, 3), dtype=np.float64, order="c")
        lambdas = np.zeros((3), dtype=np.float64, order="c")
        cppeigen(mat, dirs, lambdas)

        min_index=np.argmin(lambdas)
        if best_dir:
            dirs=[dirs[min_index]]


        # normalize result:
        dirs = np.array([self.normalize_dir(dir) for dir in dirs])

        return dirs

    def dirs_without_outliers(self, dirs, positions, op_type, use_best_dir):
        def compute_distance_from_line(group_avg_point, test_dir_end):
            def magnitude(point1, point2):
                vec = point2 - point1
                return math.sqrt(vec[0] * vec[0] + vec[1] * vec[1] + vec[2] * vec[2])

            def get_U(Point, LineStart, LineEnd, LineMag):
                return (((Point[0] - LineStart[0]) * (LineEnd[0] - LineStart[0])) +
                        ((Point[1] - LineStart[1]) * (LineEnd[1] - LineStart[1])) +
                        ((Point[2] - LineStart[2]) * (LineEnd[2] - LineStart[2]))) / (LineMag * LineMag)

            def get_intersect(LineStart, LineEnd, U):
                Intersection = [0, 0, 0]
                Intersection[0] = LineStart[0] + U * (LineEnd[0] - LineStart[0]);
                Intersection[1] = LineStart[1] + U * (LineEnd[1] - LineStart[1]);
                Intersection[2] = LineStart[2] + U * (LineEnd[2] - LineStart[2]);
                return Intersection

            linestart = [0, 0, 0]
            linemag = magnitude(test_dir_end, linestart)
            U = get_U(group_avg_point, linestart, test_dir_end, linemag)
            intersection = get_intersect(linestart, test_dir_end, U)
            return magnitude(group_avg_point, intersection)
        print("======================detecting outliers============================")
        more_dirs = []
        for dir in dirs:
            # 1.Find the distance of each point from the line/plane
            dists=[]
            for i in range(len(positions)):
                pos = positions[i]
                if op_type == 'CS':
                    dist = math.fabs(dir[0] * pos[0] + dir[1] * pos[1] + dir[2] * pos[2])
                else:
                    dist = compute_distance_from_line(pos, dir)
                dists.append(dist)

            # 2. Find median of the distances m
            median = np.median(np.array(dists))
            # 3. for each distance di, if di > 2*m, remove it as an outlier
            with_outliers_removed = list()
            for i in range(len(positions)):
                pos = positions[i]
                dist = dists[i]
                if not (dist / median > 2 or dist / median > 2):
                    with_outliers_removed.append(pos)
            # 4. recompute dirs
            outliers_dirs = self.dir_fit(with_outliers_removed, use_best_dir)
            more_dirs += list(outliers_dirs)
        return np.array(more_dirs)

    def dirs_orthogonal(self, dirs):
        added_dirs = list(dirs)

        for dir in dirs:
            if math.fabs(dir[0]) < MINDOUBLE:
                dir1 = [1.0, 0.0, 0.0]
                dir2 = [0.0, -dir[2], dir[1]]
            elif math.fabs(dir[1]) < MINDOUBLE:
                dir1 = [-dir[2], 0.0, dir[0]]
                dir2 = [0.0, 1.0, 0.0]
            else:
                dir1 = [-dir[1], dir[0], 0.0]
                dir2 = [0.0, -dir[2], dir[1]]
            # normalize dir1:
            dir1 = self.normalize_dir(dir1)
            # remove projection of dir1 from dir2
            scal = dir1[0] * dir2[0] + dir1[1] * dir2[1] + dir1[2] * dir2[2]
            dir2[0] -= scal * dir1[0]
            dir2[1] -= scal * dir1[1]
            dir2[2] -= scal * dir1[2]
            # normalize dir2
            dir2 = self.normalize_dir(dir2)
            added_dirs.append(dir1)
            added_dirs.append(dir2)

        return np.array(added_dirs)












import numpy as np
from csm.input_output.formatters import csm_log as print
from csm.calculations.basic_calculations import create_rotation_matrix, check_perm_cycles, \
    check_perm_structure_preservation
from csm.calculations.constants import MINDOUBLE, MAXDOUBLE
from csm.molecule.molecule import Molecule
from csm.molecule.normalizations import de_normalize_coords, normalize_coords
from collections import namedtuple
class CSMState(namedtuple('CSMState', ['molecule',
                                   'op_order',
                                   'op_type',
                                   'csm',
                                   'perm',
                                   'dir',
                                   'perm_count',
                                    'num_invalid'])):
    pass

CSMState.__new__.__defaults__ = (None,) * len(CSMState._fields)

class Operation:
    def __init__(self, op, sn_max=8, init=True):
        self.op_code=op
        if init:
            op=self._get_operation_data(op)
            self.type= op.type
            self.order = op.order
            if op.type=="CH":
                self.order=sn_max
            self.name = op.name


    def _get_operation_data(self, opcode):
        """
        Returns data about an operation based on the opcode
        Args:
            opcode: c2, s4, etc...

        Returns:
            And OperationCode object, with type, order and name
        """
        OperationCode = namedtuple('OperationCode', ('type', 'order', 'name'))
        _opcode_data = {
            "cs": ('CS', 2, "MIRROR SYMMETRY"),
            "ci": ('CI', 2, "INVERSION (S2)"),
            "ch": ('CH', 2, "CHIRALITY"),
            "c2": ('CN', 2, "C2 SYMMETRY"),
            'c3': ('CN', 3, "C3 SYMMETRY"),
            'c4': ('CN', 4, "C4 SYMMETRY"),
            'c5': ('CN', 5, "C5 SYMMETRY"),
            'c6': ('CN', 6, "C6 SYMMETRY"),
            'c7': ('CN', 7, "C7 SYMMETRY"),
            'c8': ('CN', 8, "C8 SYMMETRY"),
            'c10': ('CN', 10, "C10 SYMMETRY"),
            's1': ('CS', 2, "MIRROR SYMMETRY (S1)"),
            's2': ('SN', 2, "S2 SYMMETRY"),
            's4': ('SN', 4, "S4 SYMMETRY"),
            's6': ('SN', 6, "S6 SYMMETRY"),
            's8': ('SN', 8, "S8 SYMMETRY"),
            's10': ('SN', 8, "S10 SYMMETRY")
        }

        def isint(s):
            try:
                int(s)
                return True
            except ValueError:
                return False

        opcode = opcode.lower()
        if opcode[0] == 'c' and isint(opcode[1:]):
            return OperationCode(type='CN', order=int(opcode[1:]), name=opcode.upper() + ' SYMMETRY')
        if opcode[0] == 's' and isint(opcode[1:]):
            if opcode[1:] == '1':
                data = _opcode_data[opcode.lower()]
                return OperationCode(type=data[0], order=data[1], name=data[2])
            if int(opcode[1:]) % 2 != 0:
                raise ValueError("SN values must be even")
            return OperationCode(type='SN', order=int(opcode[1:]), name=opcode.upper() + ' SYMMETRY')
        try:
            data = _opcode_data[opcode.lower()]
        except KeyError:
            raise
        return OperationCode(type=data[0], order=data[1], name=data[2])



    @staticmethod
    def placeholder(op_type, op_order, sn_max=8):
        #make an arbitrary operation
        o=Operation("C2", init=False)
        #overwrite values to match input
        o.type=op_type
        o.order=op_order
        if op_type=="CH":
            o.order=sn_max
        return o

    def to_dict(self):
        return {
            "name":self.name,
            "order":self.order,
            "type":self.type
        }

    @staticmethod
    def from_dict(in_dict):
        #make an arbitrary operation
        o=Operation("C2", init=False)
        #overwrite values
        o.type=in_dict["type"]
        o.name=in_dict["name"]
        o.order=in_dict["order"]
        return o


class CSMResult:
    def __init__(self, state, operation, overall_stats={}, ongoing_stats={}):
        self.failed=False
        #input
        self.molecule=state.molecule.copy() #not yet denormalized
        self.normalized_molecule_coords = np.array(self.molecule.Q)
        self.molecule.de_normalize()
        self.operation=operation
        self.op_type=state.op_type
        self.op_order=state.op_order

        #result
        self.csm=state.csm
        self.dir=state.dir
        self.perm=state.perm
        self.normalized_symmetric_structure = self.create_symmetric_structure(self.normalized_molecule_coords, self.perm, self.dir, self.op_type,
                                                         self.op_order)
        self.symmetric_structure = de_normalize_coords(list(self.normalized_symmetric_structure), self.molecule.norm_factor)
        self.formula_csm = self.get_CSM_by_formula(self.molecule, self.symmetric_structure)

        #stats
        self.overall_statistics=overall_stats
        self.ongoing_statistics=ongoing_stats

        falsecount, num_invalid, cycle_counts, bad_indices = check_perm_cycles(self.perm, operation)
        self.overall_statistics["# bad cycles"]=falsecount
        self.overall_statistics["% bad cycles"]= num_invalid / len(self.molecule)
        try:
            self.overall_statistics["% structure"]=check_perm_structure_preservation(self.molecule, self.perm)
        except ValueError:
            self.overall_statistics["% structure"]= "n/a"

        if self.operation.name=="CHIRALITY":
            if self.op_type == 'CS':
                self.overall_statistics["best chirality"] = "CS"
            else:
                self.overall_statistics["best chirality"] =  "S%d" % (self.op_order)


        self.overall_statistics["formula CSM"]=self.formula_csm

    @property
    def d_min(self):
        return 1.0 - (self.csm / 100 * self.operation.order / (self.operation.order - 1))

    @property
    def local_csm(self):
        return self.compute_local_csm(self.molecule.Q, self.operation, self.dir)

    @property
    def chain_perm_string(self):
        molecule=self.molecule
        perm=self.perm
        chain_perm_dict = {}
        for chain in molecule.chains:
            index = molecule.chains[chain][0]
            permuted_index = perm[index]
            for chain2 in molecule.chains:
                if permuted_index in molecule.chains[chain2]:
                    chain_perm_dict[chain] = chain2
                    break
        chain_perm = []
        for chain in molecule.chains:
            permuted_index = chain_perm_dict[chain]
            chain_perm.append(permuted_index)
        chain_str=""
        for from_index, to_index in enumerate(chain_perm):
            from_chain=self.molecule.chains.index_to_string(from_index)
            to_chain=self.molecule.chains.index_to_string(to_index)
            chain_str+=from_chain + "->" + to_chain + ", "
        chain_str=chain_str[:-2] #remove final comma and space
        return chain_str

    def create_symmetric_structure(self, molecule_coords, perm, dir, op_type, op_order):
        # print('create_symmetric_structure called')

        cur_perm = np.arange(len(perm))  # array of ints...
        size = len(perm)
        m_pos = molecule_coords
        symmetric = np.copy(m_pos)

        normalization = 1 / op_order

        ########calculate and apply transform matrix#########
        ###for i<OpOrder
        for i in range(1, op_order):
            # get rotation
            rotation_matrix = create_rotation_matrix(i, op_type, op_order, dir)
            # print("Rotation matrix:\n")
            # print(rotation_matrix)
            # rotated_positions = m_pos @ rotation_matrix

            # set permutation
            cur_perm = [perm[cur_perm[j]] for j in range(size)]

            # add correct permuted rotation to atom in outAtoms
            for j in range(len(symmetric)):
                symmetric[j] += rotation_matrix @ m_pos[cur_perm[j]]
                # print("Symmetric: ", symmetric)

        # apply normalization:
        symmetric *= normalization

        return symmetric

    def get_CSM_by_formula(self, molecule, symmetric_structure):
        Q=molecule.Q
        # step one: get average of all atoms
        init_avg = np.mean(Q, axis=0)
        # step two: distance between intial and actual: initial - actual, squared
        # step three: normal: distance between initial and initial average, (x-x0)^2 + (y-y0)^2 + (z-z0)^2
        # step four: sum of distances between initial and actual, and then sum of x-y-z
        # step five: sum of normal
        distance = np.array([0.0, 0.0, 0.0])
        normal = 0.0
        for i in range(len(Q)):
            distance += (np.square(Q[i] - symmetric_structure[i]))  # square of difference
            normal += (np.sum(np.square(Q[i] - init_avg)))
        distance = np.sum(distance)
        # print("yaffa normal =", normal)
        # step six: 100 * step four / step five
        result = 100 * distance / normal
        return result

    def compute_local_csm(self, molecule_coords, operation, dir):
        size = len(molecule_coords)
        cur_perm = [i for i in range(size)]
        local_csm = np.zeros(size)
        m_pos = molecule_coords

        for i in range(operation.order):
            rot = create_rotation_matrix(i, operation.type, operation.order, dir)

            # set permutation
            cur_perm = [self.perm[cur_perm[j]] for j in range(size)]

            # apply rotation to each atoms
            rotated = rot @ m_pos[cur_perm[i]]
            difference = rotated - m_pos[i]
            square = np.square(difference)
            sum = np.sum(square)
            local_csm[i] = sum * (100.0 / (2 * operation.order))
        return local_csm

    def print_summary(self):
        try:
            percent_structure = check_perm_structure_preservation(self.molecule, self.perm)
            print("The permutation found maintains " +
                    str(round(percent_structure * 100, 2)) + "% of the original molecule's structure")

        except ValueError:
            print("The input molecule does not have bond information and therefore conservation of structure cannot be measured")

        falsecount, num_invalid, cycle_counts, bad_indices = check_perm_cycles(self.perm, self.operation)
        print(
            "The permutation found contains %d invalid %s. %.2lf%% of the molecule's atoms are in legal cycles" % (
                falsecount, "cycle" if falsecount == 1 else "cycles",
                    100 * (len(self.molecule) - num_invalid) / len(self.molecule)))

        for cycle_len in sorted(cycle_counts):
                valid = cycle_len == 1 or cycle_len == self.operation.order or (
                        cycle_len == 2 and self.operation.type == 'SN')
                count = cycle_counts[cycle_len]
                print("There %s %d %s %s of length %d" % (
                    "is" if count == 1 else "are", count, "invalid" if not valid else "",
                    "cycle" if count == 1 else "cycles",
                    cycle_len))

        if self.operation.name == "CHIRALITY":
            print("Minimum chirality was found in", self.overall_statistics["best chirality"])

        print("%s: %.4lf" % (self.operation.name, abs(self.csm)))

        print("Chain perm: " + self.chain_perm_string)


    def to_dict(self):
        return {"Result":
            {
                "molecule": self.molecule.to_dict(),
                "normalized_molecule_coords": [list(i) for i in self.normalized_molecule_coords],
                "operation": self.operation.to_dict(),

                "csm": self.csm,
                "perm": self.perm,
                "dir": list(self.dir),

                "normalized_symmetric_structure": [list(i) for i in self.normalized_symmetric_structure],
                "symmetric_structure": [list(i) for i in self.symmetric_structure],
                "formula_csm": self.formula_csm,

                "overall stats":self.overall_statistics,
                "ongoing stats":self.ongoing_statistics
            }
        }

    @staticmethod
    def from_dict():
        result_dict=input["Result"]
        molecule=Molecule.from_dict(result_dict["molecule"])
        molecule.normalize()
        operation=Operation.from_dict(result_dict["operation"])
        state=CSMState(molecule, operation.order, operation.type, result_dict["csm"], result_dict["perm"], result_dict["dir"])
        result=CSMResult(state, operation, result_dict["overall stats"], result_dict["ongoing stats"])
        return result

class FailedResult:
    def __init__(self, failed_reason, molecule, **kwargs):
        self.failed=True
        self.failed_reason=failed_reason

        self.molecule=molecule
        self.normalized_molecule_coords = []
        self.operation=kwargs["operation"]
        self.op_type=self.operation.type
        self.op_order=self.operation.order

        #result
        self.csm="n/a"
        self.dir=["n/a", "n/a", "n/a"]
        self.perm=["n/a"]
        self.normalized_symmetric_structure = []# [["n/a"] for i in range(len(molecule))]
        self.symmetric_structure = []#  [[0,0,0] for i in range(len(molecule))]
        self.formula_csm = "n/a"

        self.overall_statistics={
            "failed":"FAILED",
            "reason for failure":self.failed_reason
        }

        self.ongoing_statistics={}


"""
This module is for all the calculations that power the CSM calculation
"""
from csm.calculations.approx.approximators import ApproxCalculation
from csm.calculations.exact_calculations import ExactCalculation
from csm.calculations.trivial_calculations import TrivialCalculation
from csm.calculations.approx.dirs import DirectionChooser

Approx=ApproxCalculation
Exact=ExactCalculation
Trivial=TrivialCalculation


from shutil import copyfile

from setuptools import setup #if this import statement dooes not come first, weird errors happen
#see: https://stackoverflow.com/questions/21594925/error-each-element-of-ext-modules-option-must-be-an-extension-instance-or-2-t
from distutils.extension import Extension
import setuptools
import sys
import numpy
import os
import re
import glob
from setuptools.command.build_ext import build_ext as _build_ext

# Prepare should copy README.md to this location
with open(os.path.join(os.path.dirname(__file__), 'README.md')) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))


# Fix the numpy bootstrap problem (http://stackoverflow.com/a/21621689/871910)
class build_ext(_build_ext):
    def finalize_options(self):
        _build_ext.finalize_options(self)
        # Prevent numpy from thinking it is still in its setup process:
        __builtins__.__NUMPY_SETUP__ = False
        import numpy
        self.include_dirs.append(numpy.get_include())
        print("Include dirs are: ", self.include_dirs)

# Cython definitions
FAST_CPPUTILS_DIR = "FastCPPUtils"
EIGEN_INCLUDE_DIR_1 = "include"
EIGEN_INCLUDE_DIR_2 = "include/Eigen"

CPP_WRAPPER_DIR = "csm/CPP_wrapper"

MUNKRES_DIR = "cython-munkres"
MUNKRES_INCLUDE_DIR = "cython-munkres/src/cpp"

extra_compile_args = []
extra_link_args = []
if sys.platform == 'win32':
    extra_compile_args = ['/Ox']
    # extra_link_args = ['/debug']
elif sys.platform in ['linux', 'linux2']:
    extra_compile_args = ['-fPIC']

def get_version():
    pathname = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'csm', 'version.py')
    with open(pathname, "r") as ver_inp:
        code = ver_inp.read()
    pattern = r"__version__\s*=\s*'(?P<version>.*)'"
    match = re.match(pattern, code)
    if not match:
        raise ValueError("Version file must contain one line: __version__='..c.'")
    version = match.group("version")
    return version


class PrepareCommand(setuptools.Command):
    description = "Build fast.pyx so there's no cython dependence in installation"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        print("running prepare command")
        self.copy_md_files()
        self.copy_source_files()
        self.convert_to_c()

    def copy_md_files(self):
        # Copy LICENSE.md and README.md from our parent directory, so that setup.py sdist finds them later
        print('Copying md files from project root')
        copyfile(os.path.join('..', 'README.md'), 'README.md')
        copyfile(os.path.join('..', 'LICENSE.md'), 'LICENSE.md')

    def copy_source_files(self):
        #this may be used for copying openbabel file eventually?
        pass

    def convert_to_c(self):
        #creates fast.h and fast.c in cpp_wrapper folder
        print('Converting ProteinCSM pyx files to C++ sources...')
        pyx = './csm/CPP_wrapper/fast.pyx'
        self.cython(pyx)
        print('Converting munkres pyx files to C++ sources...')
        pyx = './cython-munkres/src/cython_munkres.pyx'
        self.cython(pyx)

    def cython(self, pyx):
        from Cython.Compiler.CmdLine import parse_command_line
        from Cython.Compiler.Main import compile
        options, sources = parse_command_line(['-2', '-v', '--cplus', pyx])
        result = compile(sources, options)
        if result.num_errors > 0:
            print('Errors converting %s to C++' % pyx, file=sys.stderr)
            raise Exception('Errors converting %s to C++' % pyx)
        self.announce('Converted %s to C++' % pyx)

csm_version = get_version()
print("Packaging ProteinCSM version %s" % csm_version)
setup(
    name='proteincsm',
    version=csm_version,
    packages=['csm.calculations', 'csm.calculations.approx', 'csm.input_output', 'csm.molecule', 'csm.main', 'csm',],
    setup_requires=['numpy>=1.10'],
    install_requires=['numpy>=1.10', 'openbabel>=2.4.0'],
    include_package_data=True,
    license='GPL2',
    description='The Continuous Symmetry Measure for Proteins',
    long_description=README,
    url='https://github.com/continuous-symmetry/proteincsm/tree/v%s' % csm_version,
    author='The Research Software Company',
    author_email='itay@chelem.co.il',

    # The csm command
    entry_points={
        'console_scripts': [
            'proteincsm = csm.main.csm_run:run_no_return',
        ]
    },

    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Topic :: Scientific/Engineering :: Chemistry',
    ],

    cmdclass={
        'prepare': PrepareCommand,
    },

    # The Cython extension module
    # We do not use Cython itself but the Cython output files
    # Note that header files that are required by the compilation are specified in MANIFEST.in
    ext_modules=[
        Extension(
            "cython_munkres",
            [os.path.join(MUNKRES_DIR, "src", "cython_munkres.cpp"),
             os.path.join(MUNKRES_DIR, "src", "cpp", "Munkres.cpp")],
            language='c++',
            include_dirs=[MUNKRES_INCLUDE_DIR, numpy.get_include()]
        ),
        Extension(
            "csm.fast",
            [os.path.join(CPP_WRAPPER_DIR, "fast.cpp"),
             os.path.join(FAST_CPPUTILS_DIR, "rpoly.c"),
             os.path.join(FAST_CPPUTILS_DIR, "math_wrappers.cpp"),
             ],
            language='c++',
            include_dirs=[EIGEN_INCLUDE_DIR_1, EIGEN_INCLUDE_DIR_2, FAST_CPPUTILS_DIR, numpy.get_include()],
            extra_compile_args=extra_compile_args,
            extra_link_args=extra_link_args),
    ],


)



# Constants used in the various calculations
from datetime import datetime

MINDOUBLE = 1e-8
MAXDOUBLE = 100000000.0
ZERO_IM_PART_MAX = 1e-3
CSM_THRESHOLD=0.0001

global global_start_time
global_start_time= datetime.now()

global global_time_out
global_time_out=50000


def set_global_timeout(timeout):
    global global_time_out
    global_time_out=timeout


import datetime
from csm.fast import CythonPermuter
from csm.calculations.data_classes import CSMState, CSMResult
from csm.calculations.constants import MAXDOUBLE
from csm.calculations.exact_calculations import ExactCalculation
from csm.molecule.molecule import MoleculeFactory
from csm.calculations.basic_calculations import run_time

'''
contains the trivial calculation (identity perm on a chain permutation)
'''

class TrivialCalculation:
    """
    Calculates the CSM of the identity permutation of a molecule. 
    If use-chains is specified, calculates the identity permutation of every possible chain permutation, returns best.
    """
    def __init__(self, operation, molecule, use_chains=True, *args, **kwargs):
        """
        :param operation: instance of Operation class or named tuple, with fields for name and order, that describes the symmetry.
        :param molecule: instance of Molecule class on which the described symmetry calculation will be performed.
        :param use_chains: default True. When True, all possible chain permutations with an identity perm on their components are measured.
                When false, only the pure identity perm is measured.
        """
        self.operation=operation
        self.molecule=molecule
        self.use_chains=use_chains
        self.statistics={}
        self.start_time=datetime.datetime.now()

    def calculate(self, *args, **kwargs):
        molecule=self.molecule
        if molecule.chains and self.use_chains:
            best = CSMState(molecule=molecule, op_type=self.operation.type, op_order=self.operation.order, csm=MAXDOUBLE)
            chain_permutations = []
            dummy = MoleculeFactory.dummy_molecule_from_size(len(molecule.chains), molecule.chain_equivalences)
            permuter = CythonPermuter(dummy, self.operation.order, self.operation.type, precalculate=False)
            for state in permuter.permute():
                chain_permutations.append(list(state.perm))
            for chainperm in chain_permutations:
                perm = [-1 for i in range(len(molecule))]
                for f_index in range(len(chainperm)):
                    f_chain = molecule.chains[f_index]
                    t_chain = molecule.chains[chainperm[f_index]]
                    for i in range(len(f_chain)):
                        perm[f_chain[i]] = t_chain[i]

                result= ExactCalculation.exact_calculation_for_approx(self.operation, molecule, perm=perm)
                if result.csm < best.csm:
                    best = result

        else:
            perm = [i for i in range(len(molecule))]
            best= ExactCalculation.exact_calculation_for_approx(self.operation, molecule, perm=perm)

        self._csm_result = CSMResult(best, self.operation, overall_stats={"runtime":run_time(self.start_time)})
        return self.result

    @property
    def result(self):
        return self._csm_result


import sys
import timeit
from csm.calculations.basic_calculations import CalculationTimeoutError
from csm.input_output.arguments import get_split_arguments
from csm.calculations import Approx, Trivial, Exact
from csm.input_output.writers import FileWriter
from csm import __version__
from csm.molecule.molecule import MoleculeReader

sys.setrecursionlimit(10000)

def run(args=[]):
    print("ProteinCSM version %s" % __version__)
    if not args:
        args = sys.argv[1:]

    #parse args
    dictionary_args = get_split_arguments(args)

    #read molecule from file
    mol=MoleculeReader.from_file(**dictionary_args)
    dictionary_args['molecule']=mol

    #print molecule printouts
    mol.print_equivalence_class_summary(dictionary_args['use_chains'])

    #call the calculation
    if dictionary_args['calc_type'] == 'approx':
        calc = Approx(**dictionary_args)
    elif dictionary_args['calc_type'] == 'trivial':
        calc = Trivial(**dictionary_args)
    else:
        csm_state_tracer_func= None
        calc=Exact(**dictionary_args, callback_func=csm_state_tracer_func)

    try:
        calc.calculate()
    except CalculationTimeoutError as e:
        print("Timed out")
        return

    result=calc.result
    #print the results
    fw=FileWriter(result, format=dictionary_args['molecule'].metadata.format, **dictionary_args)
    fw.write()
    return result

def run_no_return(args=[]):
    run(args)

if __name__ == '__main__':
    timer = timeit.Timer(lambda: run(args=sys.argv[1:]))
    time = timer.timeit(number=1)
    print("Runtime:", time, "seconds")



import json
from csm.input_output.formatters import format_CSM, non_negative_zero
import io
from openbabel import OBConversion
from csm.molecule.molecule import MoleculeReader


# molwriters
class CSMMolWriter:
    def write(self, f, result, op_name, format="csm"):
        self.print_output_csm(f, result, op_name)

    def print_output_csm(self, f, result, op_name):
        """
        Prints output in CSM format
        :param f: File to print to
        :param result: The result of the CSM calculation (a CSMState)
        :param calc_args: Calculation arguments to CSM
        """
        size = len(result.molecule.atoms)

        # print initial molecule

        f.write("\nINITIAL STRUCTURE COORDINATES\n%i\n\n" % size)
        for i in range(size):
            f.write("%3s%10.5lf %10.5lf %10.5lf\n" %
                    (result.molecule.atoms[i].symbol,
                     non_negative_zero(result.molecule.atoms[i].pos[0]),
                     non_negative_zero(result.molecule.atoms[i].pos[1]),
                     non_negative_zero(result.molecule.atoms[i].pos[2])))

        for i in range(size):
            f.write("%d " % (i + 1))
            for j in result.molecule.atoms[i].adjacent:
                f.write("%d " % (j + 1))
            f.write("\n")

        # print resulting structure coordinates

        f.write("\nMODEL 02 RESULTING STRUCTURE COORDINATES\n%i\n" % size)

        for i in range(size):
            f.write("%3s%10.5lf %10.5lf %10.5lf\n" %
                    (result.molecule.atoms[i].symbol,
                     non_negative_zero(result.symmetric_structure[i][0]),
                     non_negative_zero(result.symmetric_structure[i][1]),
                     non_negative_zero(result.symmetric_structure[i][2])))

        for i in range(size):
            f.write("%d " % (i + 1))
            for j in result.molecule.atoms[i].adjacent:
                f.write("%d " % (j + 1))
            f.write("\n")


class OBMolWriter:
    def write(self, f, result, op_name, format):
        self.print_output_ob(f, result, format, op_name)

    def print_output_ob(self, f, result, format, op_name):
        """
        Prints output using Open Babel
        :param f: File to write to
        :param result: The result of the CSM calculation (a CSMState)
        :param in_args: Input arguments to CSM
        :param calc_args: Calculation arguments to CSM
        :param out_args: Output arguments to CSM
        """
        # print initial molecule
        if format == 'pdb':
            f.write("\nMODEL 01")
        f.write("\nINITIAL STRUCTURE COORDINATES\n")

        obmol = MoleculeReader._obm_from_file(result.molecule.metadata.filepath,
                                              result.molecule.metadata.format,
                                              result.molecule.metadata.babel_bond)[0]
        for to_remove in result.molecule._deleted_atom_indices:
            obmol.DeleteAtom(obmol.GetAtom(to_remove + 1))

        num_atoms = obmol.NumAtoms()
        # update coordinates
        for i in range(num_atoms):
            try:
                atom = obmol.GetAtom(i + 1)
                atom.SetVector(non_negative_zero(result.molecule.atoms[i].pos[0]),
                               non_negative_zero(result.molecule.atoms[i].pos[1]),
                               non_negative_zero(result.molecule.atoms[i].pos[2]))
            except Exception as e:
                pass

        self.write_ob_molecule(obmol, format, f)

        # print resulting structure coordinates

        # update output coordinates to match symmetric structure
        for i in range(num_atoms):
            try:
                a = obmol.GetAtom(i + 1)
                a.SetVector(non_negative_zero(result.symmetric_structure[i][0]),
                            non_negative_zero(result.symmetric_structure[i][1]),
                            non_negative_zero(result.symmetric_structure[i][2]))
            except Exception as e:
                pass

        if format == 'pdb':
            f.write("\nMODEL 02")
        f.write("\nRESULTING STRUCTURE COORDINATES\n")
        self.write_ob_molecule(obmol, format, f)
        if format == 'pdb':
            f.write("END\n")

    def write_ob_molecule(self, mol, format, f):
        """
        Write an Open Babel molecule to file
        :param mol: The molecule
        :param format: The output format
        :param f: The file to write output to
        :param f_name: The file's name (for extension-finding purpose)
        """
        conv = OBConversion()
        if not conv.SetOutFormat(format):
            raise ValueError("Error setting output format to " + format)

        # write to file

        try:
            s = conv.WriteString(mol)
        except (TypeError, ValueError, IOError):
            raise ValueError("Error writing data file using OpenBabel")
        if str.lower(format) == 'pdb':
            s = s.replace("END", "ENDMDL")
        f.write(s)


# resultwriters
class ResultWriter:
    """
    A class for writing results. It can write molecules to various openbabel and CSM formats,
    can write headers, local csm, permutation, etc. Most functions prefixed write_ expect a filestream.
    The two print functions may be changed in the future but for now print to the screen
    Inheriting classes are recommended to inherit a write() function that can be called from main()
    """

    def __init__(self, result, op_name, format, print_local=False, *args, **kwargs):
        self.result = result
        self.op_name = op_name
        self.format = str.lower(format)
        self.print_local = print_local
        self.result_string = self.get_result_string()

    def write(self):
        raise NotImplementedError

    def get_result_string(self):
        result_io = io.StringIO()
        self._write_results(result_io)
        result_string = result_io.getvalue()
        result_io.close()
        return result_string

    def to_json(self):
        json_dict = {"Result":
            {
                "result_string": self.result_string,
                "molecule": self.result.molecule.to_json(),
                "op_order": self.result.op_order,
                "op_type": self.result.op_type,
                "csm": self.result.csm,
                "perm": self.result.perm,
                "dir": list(self.result.dir),
                "d_min": self.result.d_min,
                "symmetric_structure": [list(i) for i in self.result.symmetric_structure],
                "local_csm": self.result.local_csm,
                "perm_count": self.result.perm_count,
                "formula_csm": self.result.formula_csm,
                "normalized_molecule_coords": [list(i) for i in self.result.normalized_molecule_coords],
                "normalized_symmetric_structure": [list(i) for i in self.result.normalized_symmetric_structure],
            }
        }
        return json_dict

    def _write_results(self, f):
        self.write_header(f)
        self.write_mol(f)
        self.write_dir(f)

        if self.print_local:
            self.write_local_csm(f)

        if self.op_name == "CHIRALITY":
            self.write_chirality(f)

        self.write_permutation(f)

    def write_header(self, f):
        f.write("%s: %s\n" % (self.op_name, format_CSM(self.result.csm)))
        f.write("SCALING FACTOR: %7lf\n" % non_negative_zero(self.result.d_min))

    def write_mol(self, f):
        if self.format == "csm":
            molwriter = CSMMolWriter()
        else:
            molwriter = OBMolWriter()

        molwriter.write(f, self.result, self.op_name, self.format)

    def write_dir(self, f):
        f.write("\n DIRECTIONAL COSINES:\n\n")
        f.write("%lf %lf %lf\n" % (
            non_negative_zero(self.result.dir[0]), non_negative_zero(self.result.dir[1]),
            non_negative_zero(self.result.dir[2])))

    def write_local_csm(self, f):
        sum = 0
        f.write("\nLocal CSM: \n")
        size = len(self.result.molecule.atoms)
        for i in range(size):
            sum += self.result.local_csm[i]
            f.write("%s %7lf\n" % (self.result.molecule.atoms[i].symbol, non_negative_zero(self.result.local_csm[i])))
        f.write("\nsum: %7lf\n" % sum)

    def write_chirality(self, f):
        if self.result.op_type == 'CS':
            f.write("\n MINIMUM CHIRALITY WAS FOUND IN CS\n\n")
        else:
            f.write("\n MINIMUM CHIRALITY WAS FOUND IN S%d\n\n" % self.result.op_order)

    def write_permutation(self, f):
        f.write("\n PERMUTATION:\n\n")
        for i in self.result.perm:
            f.write("%d " % (i + 1))
        f.write("\n")

    def print_structure(self):
        self.result.print_summary()

    def print_result(self):
        print("%s: %s" % (self.op_name, format_CSM(self.result.csm)))
        print("CSM by formula: %s" % (format_CSM(self.result.formula_csm)))

    def print_chain_perm(self):
        if len(self.result.molecule.chains)>1:
            print("Chain perm: ", self.result.chain_perm_string)

class StatisticWriter:
    def __init__(self, result, stat_file_name, polar):
        try:
            self.statistics=result.statistics
        except:
            self.statistics=None
        self.file_name = stat_file_name
        self.polar = polar


    def write(self):
        if self.file_name:
            if self.statistics:
                with open(self.file_name, 'w') as file:
                    if self.polar:
                        file.write("Index"
                                "\tr_i\tth_i\tph_i"
                               "\tCSM_i"
                               "\tr_f\tth_f\tph_f"
                               "\tCSM_f"
                                "\tRuntime"
                                "\t No Iterations"
                                "\n")
                    else:
                        file.write("Index"
                                "\tx_i\ty_i\tz_i"
                               "\tCSM_i"
                               "\tx_f\ty_f\tz_f"
                               "\tCSM_f"
                                "\tRuntime"
                                "\t No Iterations"
                                "\n")
                    for key in self.statistics:
                        stat=self.statistics[key]
                        x,y,z=stat.start_dir
                        xf,yf,zf=stat.end_dir
                        if self.polar:
                            x,y,z=cart2sph(x,y,z)
                            xf,yf,zf=cart2sph(xf,yf,zf)

                        file.write(str(stat.index) +"\t"
                                   + format_CSM(x)+ "\t"+format_CSM(y)+ "\t"+format_CSM(z)+ "\t"
                                   + format_CSM(stat.start_csm)+"\t"
                                   + format_CSM(xf) + "\t" + format_CSM(yf) + "\t" + format_CSM(zf) + "\t"
                                   + format_CSM(stat.end_csm) + "\t"
                                   + str(stat.run_time) + "\t"
                                   + str(stat.num_iterations) + "\n")

class FileWriter(ResultWriter):
    """
    A ResultWriter class that writes to a file
    """

    def __init__(self, result, out_file_name, op_name, format, print_local=False, json_output=False, stat_file_name=None, polar=False, *args, **kwargs):
        self.out_file_name = out_file_name
        self.json_output = json_output
        self.statistic_writer=StatisticWriter(result, stat_file_name, polar)
        super().__init__(result, op_name, format, print_local)

    def write(self):
        self.result.print_summary()
        #self.print_structure()
        #self.print_result()
        #self.print_chain_perm()
        self.statistic_writer.write()
        if self.json_output:
            with open(self.out_file_name, 'w', encoding='utf-8') as f:
                json.dump(self.to_json(), f)
        else:
            with open(self.out_file_name, 'w', encoding='utf-8') as f:
                self._write_results(f)


class FolderWriter(ResultWriter):
    pass


def print_results(result, dictionary_args):
    """
    Prints the CSM calculation results
    :param result:
    :param dictionary_args:
    """
    filewriter = FileWriter(result, **dictionary_args)
    filewriter.write()



'''
the classes used for running the approximate algorithm
'''

import datetime
from collections import OrderedDict

import numpy as np
from csm.fast import CythonPermuter
from csm.fast import approximate_perm_classic, munkres_wrapper

from csm.calculations.approx.dirs import ClassicDirectionChooser
from csm.calculations.basic_calculations import create_rotation_matrix, array_distance, check_perm_cycles, \
    CalculationTimeoutError, check_timeout
from csm.calculations.basic_calculations import now, run_time
from csm.calculations.constants import MAXDOUBLE, CSM_THRESHOLD, MINDOUBLE
from csm.calculations.data_classes import CSMState, Operation, CSMResult
from csm.calculations.exact_calculations import ExactCalculation
from csm.input_output.formatters import csm_log as print
from csm.input_output.formatters import format_CSM
from csm.molecule.molecule import MoleculeFactory


class _OptionalLogger:
    '''
    Utility class for having calls to a function which, if non-existent, does nothing
    '''

    def __init__(self, log_func=None):
        self._log_func = log_func

    def _log(self, *args):
        if self._log_func:
            self._log_func(*args)


class _SingleDirectionStatistics:
    # per direction, we want to store:
    # 1. every direction passed through
    # 2. every csm passed through, and percent cycle preservation
    # 3. runtime
    def __init__(self, dir):
        self.start_dir = dir
        self.dirs = []
        self.csms = []
        self.cycle_stats = []
        self._stop_reason = ""

    def append_sub_direction(self, result):
        self.dirs.append(result.dir)
        self.csms.append(result.csm)
        self.cycle_stats.append(result.num_invalid)

    @property
    def stop_reason(self):
        return self._stop_reason

    @stop_reason.setter
    def stop_reason(self, reason):
        self._stop_reason = reason

    def start_clock(self):
        self.__start_time = datetime.datetime.now()

    def end_clock(self):
        now = datetime.datetime.now()
        time_d = now - self.__start_time
        self.run_time = time_d.total_seconds()

    def __repr__(self):
        return str({
            "dirs": self.dirs,
            "csms": self.csms
        })

    @property
    def end_dir(self):
        return self.dirs[-1]

    @property
    def start_csm(self):
        return self.csms[0]

    @property
    def end_csm(self):
        return self.csms[-1]

    @property
    def num_iterations(self):
        return len(self.dirs)

    def __lt__(self, other):
        try:  # iof other doesn't have csm, we are less than them
            that_one = other.end_csm
        except:
            return True

        try:  # if we don't have csm, other is less than us
            this_one = self.end_csm
        except:
            return False

        return self.end_csm < other.end_csm

    def to_dict(self):
        try:
            return_dict = {
                "start dir": list(self.start_dir),
                "start csm": self.start_csm,
                "stop reason": self.stop_reason,
                "end dir": list(self.end_dir),
                "end csm": self.end_csm,
                "num iterations": self.num_iterations,
                "dirs": [list(dir) for dir in self.dirs],
                "csms": self.csms,
                "cycle stats": self.cycle_stats,
                "run time": self.run_time
            }
            return return_dict
        except:
            return {
                "start dir": list(self.start_dir),
                "stop reason": "was never reached"
            }


class DirectionStatisticsContainer:
    def __init__(self, initial_directions):
        self.directions_dict = OrderedDict()
        self.directions_arr = []
        for index, dir in enumerate(initial_directions):
            self.directions_dict[tuple(dir)] = _SingleDirectionStatistics(dir)
            self.directions_arr.append(self.directions_dict[tuple(dir)])

    def __getitem__(self, key):
        return self.directions_dict[tuple(key)]

    def __setitem__(self, key, value):
        self.directions_dict[tuple(key)] = value

    def __iter__(self):
        return self.directions_dict.__iter__()

    def __str__(self):
        return str(self.directions_dict)

    def to_dict(self):
        return [{"dir": dir, "stats": self.directions_dict[dir].to_dict()} for dir in self.directions_dict]


class ApproxStatistics(DirectionStatisticsContainer):
    pass


class SingleDirApproximator(_OptionalLogger):
    def __init__(self, operation, molecule, perm_from_dir_builder, log_func=None, timeout=100,
                 max_iterations=50):
        self._log_func = log_func
        self._molecule = molecule
        self._op_type = operation.type
        self._op_order = operation.order
        self._operation = operation
        self.max_iterations = max_iterations
        self.perm_from_dir_builder = perm_from_dir_builder(operation, molecule, log_func, timeout)
        self._chain_permutations = self.perm_from_dir_builder.get_chain_perms()
        self.timeout = timeout
        self.start = datetime.datetime.now()

    def _create_perm_from_dir(self, dir, chainperm):
        return self.perm_from_dir_builder.create_perm_from_dir(dir, chainperm)

    def calculate(self, dir):
        statistics = _SingleDirectionStatistics(dir)
        statistics.start_clock()
        best = CSMState(molecule=self._molecule, op_type=self._op_type, op_order=self._op_order, csm=MAXDOUBLE)
        self.least_invalid = CSMState(molecule=self._molecule, op_type=self._op_type, op_order=self._op_order,
                                      csm=MAXDOUBLE, num_invalid=MAXDOUBLE)
        self._log("Calculating for initial direction: ", dir)
        for chainperm in self._chain_permutations:
            if len(self._chain_permutations) > 1:
                self._log("\tCalculating for chain permutation ", chainperm)
            best_for_chain_perm = old_results = CSMState(molecule=self._molecule, op_type=self._op_type,
                                                         op_order=self._op_order,
                                                         csm=MAXDOUBLE, dir=dir)
            i = 0
            while True:
                i += 1

                self._log("\t\titeration", i, ":")

                try:
                    perm = self._create_perm_from_dir(old_results.dir, chainperm)
                    interim_results = ExactCalculation.exact_calculation_for_approx(self._operation,
                                                                                    self._molecule, perm=perm)
                except CalculationTimeoutError as e:
                    self._log("\t\titeration ", i, " Timed out after ", str(e.timeout_delta), " seconds")
                    break

                if interim_results.num_invalid < self.least_invalid.num_invalid or \
                        (interim_results.num_invalid == self.least_invalid.num_invalid
                         and interim_results.csm < self.least_invalid.csm):
                    self.least_invalid = interim_results
                statistics.append_sub_direction(interim_results)
                # self._log("\t\t\tfound a permutation using dir", old_results.dir, "...")
                if i > 1:
                    self._log("\t\t\tthere are",
                              len(perm) - np.sum(np.array(perm) == np.array(old_results.perm)),
                              "differences between new permutation and previous permutation")
                self._log("\t\t\tthe csm for this permutation is:", str(round(interim_results.csm, 8)))
                self._log("\t\t\tthe new direction from this permutation is", interim_results.dir)
                self._log("\t\t\tthe distance between the new direction and the previous direction is:",
                          str(round(np.linalg.norm(interim_results.dir - old_results.dir), 8)))

                if interim_results.csm < best_for_chain_perm.csm:
                    best_for_chain_perm = interim_results

                # Various stop conditions for the loop, listed as multiple if statements so that the code is clearer
                if i >= self.max_iterations:
                    statistics.stop_reason = "Max iterations"
                    self._log("\t\tStopping after %d iterations" % i)
                    break
                # if i > 1 and math.fabs(old_results.csm - interim_results.csm) / math.fabs(old_results.csm) > 0.01:
                #    self._log("\t\tStopping due to CSM ratio")
                if best_for_chain_perm.csm < CSM_THRESHOLD:
                    statistics.stop_reason = "CSM below threshold"
                    self._log("\t\tStopping because the best CSM is good enough")
                    break
                if abs(np.linalg.norm(interim_results.dir - old_results.dir)) <= 0:
                    statistics.stop_reason = "No change in direction"
                    self._log("\t\tStopping because the direction has not changed")
                    break
                if interim_results.csm >= old_results.csm:  # We found a worse CSM
                    statistics.stop_reason = "No improvement in CSM"
                    self._log("\t\tStopping because CSM did not improve (worse or equal)")
                    break

                old_results = interim_results

            if best_for_chain_perm.csm < best.csm:
                best = best_for_chain_perm
                if best_for_chain_perm.csm < CSM_THRESHOLD:
                    break

        statistics.end_clock()
        return best, statistics


class _PermFromDirBuilder(_OptionalLogger):
    def __init__(self, operation, molecule, log_func, timeout):
        self.start_time = datetime.datetime.now()
        self._log_func = log_func
        self._molecule = molecule
        self.operation = operation
        self._op_type = operation.type
        self._op_order = operation.order
        self.timeout = timeout
        self._precalculate()

    def _precalculate(self):
        '''should be overridden by inheriting classes that need chain permutations'''
        self._chain_permutations = [[0]]

    def get_chain_perms(self):
        return self._chain_permutations

    def create_perm_from_dir(self, dir, chainperm):
        raise NotImplementedError


class _ChainPermsPermBuilder(_PermFromDirBuilder):
    def _calc_chain_permutations(self):
        chain_permutations = []
        dummy = MoleculeFactory.dummy_molecule_from_size(len(self._molecule.chains), self._molecule.chain_equivalences)
        permuter = CythonPermuter(dummy, self._op_order, self._op_type, precalculate=False)
        for state in permuter.permute():
            chain_permutations.append([i for i in state.perm])
        return chain_permutations

    def _precalculate(self):
        self._chain_permutations = self._calc_chain_permutations()


class _GreedyPermBuilder(_ChainPermsPermBuilder):
    '''
    This uses the Cython implementation of the classic (greedy) approximate algorithm.
    It is not optimized for molecules with many chain permutations.
    '''

    def create_perm_from_dir(self, dir, chainperm):
        return approximate_perm_classic(self._op_type, self._op_order, self._molecule, dir, chainperm)


class _HungarianPermBuilder(_ChainPermsPermBuilder):
    '''
    This uses the Hungarian (munkres) algorithm for optimization of cost matrix.
         It is not optimized for molecules with many chain permutations.
    '''

    def create_perm_from_dir(self, dir, chainperm):
        perm = self.approximate_perm_hungarian(self._op_type, self._op_order, self._molecule, dir, chainperm)
        perm = self.cookie_dough(perm)
        return perm

    class DistanceMatrix:

        def __init__(self, group_size):
            self.group_size = group_size
            # ##print("Creating DistanceMatrix for group of size ", self.group_size)
            self.mv_distances = np.ones((group_size, group_size), order="c") * MAXDOUBLE
            self._allowed_rows = np.zeros(group_size, dtype='long')
            self._allowed_cols = np.zeros(group_size, dtype='long')
            # ##print("DistanceMatrix created")

        def add(self, from_val, to_val, distance=MAXDOUBLE):
            self.mv_distances[from_val, to_val] = distance
            self._allowed_rows[from_val] = 1
            self._allowed_cols[to_val] = 1

        def remove(self, from_val, to_val):
            self._allowed_rows[from_val] = 0
            self._allowed_cols[to_val] = 0

        def get_min_val(self):
            min = MAXDOUBLE
            min_i = -1
            min_j = -1
            allowed_rows = self._allowed_rows[0]
            allowed_cols = self._allowed_cols[0]

            for i in range(self.group_size):
                if self._allowed_rows[i]:
                    row_ptr = self.mv_distances[i, 0]
                    for j in range(self.group_size):
                        if self._allowed_cols[j]:
                            tmp = row_ptr[j]
                            if tmp < min:
                                min = tmp
                                min_i = i
                                min_j = j
            return (min_i, min_j)

        def get_next_in_cycle(self, from_val, constraints):
            if not self._allowed_rows[from_val]:
                self.tostr()
                raise ValueError("get_next_in_cycle called with an unavailable row %d" % from_val)

            searched_row = self.mv_distances[from_val]
            min = MAXDOUBLE
            min_i = -1
            for i in range(self.group_size):
                if self._allowed_cols[i] and not i in constraints:
                    tmp = searched_row[i]
                    if tmp < min:
                        min, min_i = tmp, i

            if min_i == -1:
                # ##print("Can't find next in cycle. Allowed columns:")
                # for i in range(self.group_size):
                #   if self._allowed_cols[i]:
                #        ##print(i, '-->', searched_row[i])
                self.tostr()
                raise ValueError("Can't find next in cycle. Constraints: %s" % str(constraints))
            return (from_val, min_i)

        def get_matrix(self):
            return self.mv_distances

        def tostr(self):
            ##print(self.mv_distances.base)
            pass

    def cycle_builder(self, chainperm):
        """
        the only thing this function does is build cycles ON THE BASIS OF CHAINPERM
        if chainperm has invalid cycles, this will return invalid cycles
        :param chainperm:
        :return: a cycle in chainperm
        """
        chainperm_copy = list(chainperm)

        ##print("chainperm_copy", chainperm_copy)

        def recursive_cycle_builder(chain_head, index, cycle):
            ##print(chain_head, index, cycle)
            cycle.append(index)
            chainperm_copy[index] = -1
            if chainperm[index] == chain_head:
                ##print("yielding cycle", cycle)
                yield cycle
                cycle = []
            else:
                yield from recursive_cycle_builder(chain_head, chainperm[index], cycle)

        for i, val in enumerate(chainperm_copy):
            ##print("I,val", i,val)
            if val == -1:
                continue
            chainperm_copy[i] = -1
            ##print("cycle head is", i)
            yield from recursive_cycle_builder(i, i, [])
        # find and return cycles
        pass

    def get_atom_indices(self, cycle, chain_group):
        indices = []
        for chain_index in cycle:
            indices += chain_group[chain_index]
        return indices

    def get_atom_to_matrix_indices(self, atom_indices, atom_to_matrix_indices):
        i = 0
        for index in atom_indices:
            atom_to_matrix_indices[index] = i
            i += 1
        return atom_to_matrix_indices

    def fill_distance_matrix(self, len_group, cycle, chain_group, chain_perm, rotated_holder, Q_holder, matrix_indices):
        distances = self.DistanceMatrix(len_group)

        for chain_index in cycle:  # Use an iterator
            # Todo: Convert from_chain and to_chain into vector[ints]
            from_chain = chain_group[chain_index]
            to_chain = chain_group[chain_perm[chain_index]]
            for i in range(len(from_chain)):
                from_chain_index = from_chain[i]
                for j in range(len(to_chain)):
                    to_chain_index = to_chain[j]
                    a = rotated_holder[from_chain_index]
                    b = Q_holder[to_chain_index]
                    distance = array_distance(a, b)
                    distances.add(matrix_indices[to_chain_index], matrix_indices[from_chain_index], distance)
        return distances

    def hungarian_perm_builder(self, op_type, op_order, group, distance_matrix, perm):
        matrix = distance_matrix.get_matrix()
        # m = Munkres()
        # indexes = m.compute(matrix)
        indexes = munkres_wrapper(matrix)
        for (from_val, to_val) in indexes:
            perm[group[from_val]] = group[to_val]
        return perm

    def approximate_perm_hungarian(self, op_type, op_order, molecule, dir, chain_perm):
        # print("Inside estimate_perm, dir=%s" % dir)
        # create rotation matrix
        rotation_mat = create_rotation_matrix(1, op_type, op_order, dir)
        # run rotation matrix on atoms
        rotated = (rotation_mat @ molecule.Q.T).T
        atom_to_matrix_indices = np.ones(len(molecule), dtype='long') * -1
        # empty permutation:
        perm = [-1] * len(molecule)

        # permutation is built by "group": equivalence class, and valid cycle within chain perm (then valid exchange w/n cycle)
        for cycle in self.cycle_builder(chain_perm):
            # Todo: Convert cycle into a vector[int]
            for chains_in_group in molecule.groups_with_internal_chains:
                # 1. create the group of atom indices we will be building a distance matrix with
                try:
                    current_atom_indices = self.get_atom_indices(cycle, chains_in_group)
                except KeyError:  # chaingroup does not have chains belonging to current cycle
                    continue  # continue to next chain group
                atom_to_matrix_indices = self.get_atom_to_matrix_indices(current_atom_indices, atom_to_matrix_indices)

                # 2. within that group, go over legal switches and add their distance to the matrix
                distances = self.fill_distance_matrix(len(current_atom_indices), cycle, chains_in_group, chain_perm,
                                                      rotated,
                                                      molecule.Q, atom_to_matrix_indices)

                # 3. call the perm builder on the group
                perm = self.hungarian_perm_builder(op_type, op_order, current_atom_indices, distances, perm)
                # (4. either continue to next group or finish)
        # print(perm)
        return perm

    def cookie_dough(self, perm):
        falsecount, num_invalid, cycle_counts, indices_in_bad_cycles = check_perm_cycles(perm, self.operation)
        return perm


class _ManyChainsPermBuilder(_PermFromDirBuilder):
    '''
    This approximator uses the Hungarian (munkwres) algorithm to choose the optimal permutation of chains, rather than
    iterating through all possible chain permutations. It is hence more efficient for molecules with many possible chain
    permutations
    '''

    def create_perm_from_dir(self, dir, chainperm="dont care"):

        chain_len = len(self._molecule.chains[0])
        for chain in self._molecule.chains:
            test_len = len(self._molecule.chains[chain])
            if test_len != chain_len:
                raise ValueError("--many-chains currently expects all chains to be of same length")

        rotation_mat = create_rotation_matrix(1, self._op_type, self._op_order, dir)
        perm = [-1] * len(self._molecule)
        # improved use chains algorithm:
        # for a given symmetric axis:
        # M= number of fragments in molecule. 0>i>M, 0>j>M
        num_fragments = (len(self._molecule.chains))
        # A1: matrix A of size MxM: initialize with zeros
        fragment_distance_matrix = np.ones((num_fragments, num_fragments), order="c") * MAXDOUBLE
        # A2: confirm that there's same number of atoms in each equivalence group between fragment and fragment j
        # if there isn't, we leave M[i,j]=0 and continue to next ij
        # (this is already confirmed in molecule setup, so we will use the equivalent chains from mol.chain_equivalences
        for equivalent_chain_group in self._molecule.chain_equivalences:
            for i, frag_i in enumerate(equivalent_chain_group):
                for j, frag_j in enumerate(equivalent_chain_group):
                    fragment_distance_matrix[i, j] = self._get_fragment_distance(frag_i, frag_j, rotation_mat)

        # Run the hungarian algorithm on Aij, and thereby find a permutation between the fragments
        indexes = munkres_wrapper(fragment_distance_matrix)
        # C: rerun A3 on the relevant ijs
        for (i, j) in indexes:
            frag_i_groups = self._molecule.chains_with_internal_groups[i]
            frag_j_groups = self._molecule.chains_with_internal_groups[j]
            # D: put together into a full permutation
            for k, group_k in enumerate(frag_i_groups):
                group_m = frag_j_groups[k]
                indexes, group_distance_matrix = self._hungarian_on_groups(group_k, group_m, rotation_mat)
                for (from_val, to_val) in indexes:
                    perm[group_k[from_val]] = group_m[to_val]
        return perm

    def _get_fragment_distance(self, frag_i, frag_j, rotation_mat):
        total_distance = 0
        frag_i_groups = self._molecule.chains_with_internal_groups[frag_i]
        frag_j_groups = self._molecule.chains_with_internal_groups[frag_j]
        # A3: e = number of equivalence groups in fragment i, of size N1... Ne
        # 0>k>e
        for k, group_k in enumerate(frag_i_groups):
            group_m = frag_j_groups[k]
            indexes, group_distance_matrix = self._hungarian_on_groups(group_k, group_m, rotation_mat)
            # and the result (=sum of matrix members on diagonal generalized that hungarian found) ????
            # we add to A[i,j]
            for (i, j) in indexes:
                total_distance += group_distance_matrix[i, j]
        return total_distance

    def _hungarian_on_groups(self, group_k, group_m, rotation_mat):
        # matrix B of size Nk x Nk
        if not group_k:
            return [], []  # the way we built chains with internal groups, groups with no members are None
        size = len(group_k)
        group_distance_matrix = np.ones((size, size), order="c") * MAXDOUBLE
        # Vi..Vnk are the coordinate vectors of fragment i in equivalence class k
        # Wi..Wnk are the coordinate vectors of fragment j in equivalence class k
        for a, index_v in enumerate(group_k):
            for b, index_w in enumerate(group_m):
                coord_v = self._molecule.Q[index_v]
                coord_w = self._molecule.Q[index_w]
                # T is the symmetric translation appropriate to the given symmetric axis
                translated_v = rotation_mat @ coord_v
                # in position a,b of matrix B we place the square of the distance between T(Va) and Wb
                # in other words matrix B is a single block from the standard approx algorithm's distance matrix,
                # specifically the block that matches fragments i, j and equivalence group k
                distance = array_distance(translated_v, coord_w)
                group_distance_matrix[a, b] = distance
        # B: We run the hungarian algorithm on matrix B,
        indexes = munkres_wrapper(group_distance_matrix)
        return indexes, group_distance_matrix


class ApproxCalculation(_OptionalLogger):
    def __init__(self, operation, molecule, approx_algorithm='greedy',
                 log_func=lambda *args: None, selective=False, num_selected=10, *args, **kwargs):
        self.operation=operation
        self._molecule = molecule

        self._log_func = log_func

        # choose the approximator class
        if approx_algorithm == 'hungarian':
            self.perm_builder = _HungarianPermBuilder
        if approx_algorithm == 'greedy':
            self.perm_builder = _GreedyPermBuilder
        if approx_algorithm == 'many-chains':
            self.perm_builder = _ManyChainsPermBuilder

        # get the directions
        direction_chooser=ClassicDirectionChooser(molecule, operation.type, operation.order)
        self._initial_directions = direction_chooser.dirs
        if len(self._initial_directions) > 1:
            self._log("There are", len(self._initial_directions),
                      "initial directions to search for the best permutation")


        self.selective = selective
        self.num_selected = num_selected
        self.statistics = ApproxStatistics(self._initial_directions)
        self._max_iterations = 30

    def calculate(self, timeout=100, *args, **kwargs):
        self.start_time = now()
        self.timeout = timeout
        overall_stats = {}
        if self.operation.name == "CHIRALITY":  # Chirality
            # First CS
            best_op = Operation('cs')
            best_result = self._calculate(best_op)
            if best_result.csm > MINDOUBLE:
                # Try the SN's
                for op_order in range(2, self.operation.order + 1, 2):
                    op = Operation("S" + str(op_order))
                    result = self._calculate(op)
                    if result.csm < best_result.csm:
                        best_result = result
                        best_op = op
                    if best_result.csm < MINDOUBLE:
                        break
        else:
            best_result = self._calculate(self.operation)
        overall_stats["runtime"] = run_time(self.start_time)
        self.result = CSMResult(best_result, self.operation, overall_stats=overall_stats,
                                ongoing_stats={"approx": self.statistics.to_dict()})
        return self.result

    def _calculate(self, operation):
        if operation.type == 'CI' or (operation.type == 'SN' and operation.order == 2):
            dir = [1.0, 0.0, 0.0]
            if operation.type == 'SN':
                op_msg = 'S2'
            else:
                op_msg = 'CI'
            self._log("Operation %s - using just one direction: %s" % (op_msg, dir))
            best_results = self._calculate_for_directions(operation, [dir], 1)
        else:
            if self.selective:
                self._calculate_for_directions(operation, self._initial_directions, 1)
                best_dirs = []
                sorted_csms = sorted(self.statistics.directions_arr)
                for item in sorted_csms[:self.num_selected]:
                    best_dirs.append(item.start_dir)
                    self._log("Running again on the", self.num_selected, "best directions")
                best_results = self._calculate_for_directions(operation, best_dirs, self._max_iterations)

            else:
                best_results = self._calculate_for_directions(operation, self._initial_directions, self._max_iterations)

        best_result, least_invalid = best_results
        if least_invalid.num_invalid < best_result.num_invalid:
            if least_invalid.csm <= best_result.csm:
                best_result = least_invalid
            else:
                print("(A result with better preservation of integrity of cycle lengths was found")
                print("Direction: ", least_invalid.dir, " yields a CSM of", format_CSM(least_invalid.csm),
                      "\n", (1 - (least_invalid.num_invalid / len(self._molecule))) * 100,
                      "% of the molecule's atoms are in legal cycles)")
        return best_result

    def _calculate_for_directions(self, operation, dirs, max_iterations):
        best = CSMState(molecule=self._molecule, op_type=operation.type, op_order=operation.order, csm=MAXDOUBLE,
                        num_invalid=MAXDOUBLE)
        least_invalid = CSMState(molecule=self._molecule, op_type=operation.type, op_order=operation.order,
                                 csm=MAXDOUBLE, num_invalid=MAXDOUBLE)
        single_dir_approximator = SingleDirApproximator(operation, self._molecule,
                                                        self.perm_builder, self._log,
                                                        self.timeout, max_iterations=max_iterations)
        for dir in dirs:
            best_result_for_dir, statistics = single_dir_approximator.calculate(dir)
            self.statistics[dir] = statistics
            least_invalid_for_dir = single_dir_approximator.least_invalid
            if least_invalid_for_dir.num_invalid < least_invalid.num_invalid or \
                    (least_invalid_for_dir.num_invalid == least_invalid.num_invalid and least_invalid_for_dir.csm < least_invalid.csm):
                least_invalid = least_invalid_for_dir
            if best_result_for_dir.csm < best.csm:
                best = best_result_for_dir
                if best.csm < CSM_THRESHOLD:
                    break
        return best, least_invalid
